import os
import pybind11
import importlib
import sysconfig
import subprocess
from functools import partial, reduce

import jax
import jax.numpy as jnp
from jax import core, dtypes
from jax.core import ShapedArray
from jax.interpreters import mlir, xla
from jax.interpreters.mlir import ir
from jax.lib import xla_client
from jaxlib.hlo_helpers import custom_call


use_rocm = os.environ.get("RWKV_USE_ROCM", "0") == "1"
if use_rocm:
    kernel_dir_name = "jax_kernel_hip"
    cuda_lib_dir = "/opt/rocm/include"
    kernel_name = "gpu_ops"
    platform = "ROCM"
else:
    kernel_dir_name = "jax_kernel_cuda"
    cuda_lib_dir = "/usr/local/cuda/include"
    kernel_name = "gpu_ops"
    platform = "gpu"


def default_layouts(*shapes):
    return [range(len(shape) - 1, -1, -1) for shape in shapes]


class RWKVKernelOperator:
    def __init__(self, head_size, max_sequence_length):
        """
        加载或构建rwkv内核
        """
        rwkv_kernel = RWKVKernelOperator._load_or_build_kernel(
            head_size, max_sequence_length
        )
        """
        向mlir注册C++算子入口
        """
        for _name, _value in rwkv_kernel.get_rwkv_registrations().items():
            xla_client.register_custom_call_target(_name, _value, platform=platform)

        """
        构建jax前端数据类型到C++后端的映射
        """

        def element_type_to_descriptor_type_mapping(element_type):
            _element_type_to_descriptor_type_mapping = {
                ir.BF16Type.get(): rwkv_kernel.ElementType.BF16,
                ir.F16Type.get(): rwkv_kernel.ElementType.F16,
                ir.F32Type.get(): rwkv_kernel.ElementType.F32,
            }
            return _element_type_to_descriptor_type_mapping.get(element_type)

        """
        定义前向过程算子
        """

        def build_forward_op():
            _rwkv_fwd_p = core.Primitive("rwkv_fwd")
            _rwkv_fwd_p.multiple_results = False
            _rwkv_fwd_p.def_impl(partial(xla.apply_primitive, _rwkv_fwd_p))

            """
            构建前向过程方法
            """

            def _rwkv_fwd(r, k, v, w, u):
                y = _rwkv_fwd_p.bind(r, k, v, w, u)
                ctx = r, k, v, w, u
                return y, ctx

            """
            milr调用上面注册的前向过程算子，milr帮助算子创建输出张量
            """

            def _rwkv_fwd_cuda_lowering(ctx, r, k, v, w, u):
                r_type = ir.RankedTensorType(r.type)
                k_type = ir.RankedTensorType(k.type)
                v_type = ir.RankedTensorType(v.type)
                w_type = ir.RankedTensorType(w.type)
                u_type = ir.RankedTensorType(u.type)
                assert all(
                    [
                        r_type.element_type == xx.element_type
                        for xx in [k_type, v_type, w_type, u_type]
                    ]
                )
                assert all(
                    [r_type.shape == xx.shape for xx in [k_type, v_type, w_type]]
                )
                assert r_type.element_type in [
                    ir.F32Type.get(),
                    ir.BF16Type.get(),
                    ir.F16Type.get(),
                ]
                bz, seq_len, hd_sz = r_type.shape

                assert hd_sz % head_size == 0
                assert reduce(lambda x, y: x * y, u_type.shape, 1) == hd_sz, (
                    "the elements of u (time first) is not equal to hidden_size"
                )
                input_type = r_type.element_type

                if input_type in [ir.F32Type.get(), ir.BF16Type.get()]:
                    output_type = input_type
                else:
                    output_type = ir.F32Type.get()

                opaque = rwkv_kernel.create_rwkv_descriptor(
                    bz,
                    seq_len,
                    hd_sz,
                    hd_sz // head_size,
                    False,  # unused
                    element_type_to_descriptor_type_mapping(input_type),
                    element_type_to_descriptor_type_mapping(output_type),
                )

                out = custom_call(
                    b"wkv_forward",
                    result_types=[
                        ir.RankedTensorType.get(r_type.shape, output_type),
                    ],
                    operands=[r, k, v, w, u],
                    backend_config=opaque,
                    operand_layouts=default_layouts(
                        r_type.shape,
                        k_type.shape,
                        v_type.shape,
                        w_type.shape,
                        u_type.shape,
                    ),
                    result_layouts=default_layouts(r_type.shape),
                ).results
                return out

            """
            将算子绑定到C++
            """
            mlir.register_lowering(
                _rwkv_fwd_p,
                _rwkv_fwd_cuda_lowering,
                platform="gpu",
            )
            """
            定义抽象过程，告知jax输出张量形状与数据类型
            """

            def _rwkv_fwd_abstract(r, k, v, w, u):
                assert all([r.shape == xx.shape for xx in [k, v, w]])
                assert len(r.shape) == 3
                bz, seq_len, channels = r.shape
                assert channels % head_size == 0
                assert seq_len <= max_sequence_length
                assert reduce(lambda x, y: x * y, u.shape, 1) == channels, (
                    "the elements of u (time first) is not equal to hidden_size"
                )

                r_dtype = dtypes.canonicalize_dtype(r.dtype)
                k_dtype = dtypes.canonicalize_dtype(k.dtype)
                v_dtype = dtypes.canonicalize_dtype(v.dtype)
                w_dtype = dtypes.canonicalize_dtype(w.dtype)
                u_dtype = dtypes.canonicalize_dtype(u.dtype)

                assert all(
                    [r_dtype == xx for xx in [k_dtype, v_dtype, w_dtype, u_dtype]]
                )
                assert r_dtype in [jnp.float32, jnp.float16, jnp.bfloat16]
                if r_dtype in [jnp.float32, jnp.bfloat16]:
                    output_dtype = r_dtype
                else:
                    output_dtype = jnp.float32
                return ShapedArray(
                    r.shape, output_dtype, named_shape=r.named_shape
                )  # output

            _rwkv_fwd_p.def_abstract_eval(_rwkv_fwd_abstract)
            return _rwkv_fwd, _rwkv_fwd_p

        rwkv_fwd, _ = build_forward_op()

        def build_backward_op():
            """
            注册反向传播过程算子
            """
            _rwkv_bwd_p = core.Primitive("rwkv_bwd")
            _rwkv_bwd_p.multiple_results = True
            _rwkv_bwd_p.def_impl(partial(xla.apply_primitive, _rwkv_bwd_p))

            """
                反向传播过程
            """

            def _rwkv_bwd(ctx, gy):
                r, k, v, w, u = ctx
                gr, gk, gv, gw, gu = _rwkv_bwd_p.bind(r, k, v, w, u, gy)
                gu = jnp.sum(gu, axis=0)

                return gr, gk, gv, gw, gu

            """
                反向传播mlir后端
            """

            def _rwkv_bwd_cuda_lowering(ctx, r, k, v, w, u, gy):
                r_type = ir.RankedTensorType(r.type)
                k_type = ir.RankedTensorType(k.type)
                v_type = ir.RankedTensorType(v.type)
                w_type = ir.RankedTensorType(w.type)
                u_type = ir.RankedTensorType(u.type)
                gy_type = ir.RankedTensorType(gy.type)

                assert all(
                    [
                        r_type.element_type == xx.element_type
                        for xx in [k_type, v_type, w_type, u_type]
                    ]
                )
                assert all(
                    [
                        r_type.shape == xx.shape
                        for xx in [k_type, v_type, w_type, gy_type]
                    ]
                )
                assert r_type.element_type in [
                    ir.F32Type.get(),
                    ir.BF16Type.get(),
                    ir.F16Type.get(),
                ]
                bz, seq_len, hd_sz = r_type.shape

                assert hd_sz % head_size == 0
                assert reduce(lambda x, y: x * y, u_type.shape, 1) == hd_sz, (
                    "the elements of u (time first) is not equal to hidden_size"
                )
                input_type = r_type.element_type

                if input_type in [ir.F32Type.get(), ir.BF16Type.get()]:
                    output_type = input_type
                else:
                    output_type = ir.F32Type.get()

                assert output_type == gy_type.element_type

                opaque = rwkv_kernel.create_rwkv_descriptor(
                    bz,
                    seq_len,
                    hd_sz,
                    hd_sz // head_size,
                    False,  # unused
                    element_type_to_descriptor_type_mapping(input_type),
                    element_type_to_descriptor_type_mapping(output_type),
                )

                gu_shape = (bz,) + u_type.shape

                out = custom_call(
                    b"wkv_backward",
                    result_types=[
                        ir.RankedTensorType.get(r_type.shape, output_type),  # gr
                        ir.RankedTensorType.get(k_type.shape, output_type),  # gk
                        ir.RankedTensorType.get(v_type.shape, output_type),  # gw
                        ir.RankedTensorType.get(w_type.shape, output_type),  # gw
                        ir.RankedTensorType.get(gu_shape, output_type),  # gu
                    ],
                    operands=[r, k, v, w, u, gy],
                    backend_config=opaque,
                    operand_layouts=default_layouts(
                        r_type.shape,
                        k_type.shape,
                        v_type.shape,
                        w_type.shape,
                        u_type.shape,
                        gy_type.shape,
                    ),
                    result_layouts=default_layouts(
                        r_type.shape, k_type.shape, v_type.shape, w_type.shape, gu_shape
                    ),
                ).results
                return out

            """
            注册反向传播算子到mlir
            """
            mlir.register_lowering(
                _rwkv_bwd_p,
                _rwkv_bwd_cuda_lowering,
                platform="gpu",
            )
            """
            反向传播的抽象过程
            """

            def _rwkv_bwd_abstract(r, k, v, w, u, gy):
                assert all([r.shape == xx.shape for xx in [k, v, w]])
                assert len(r.shape) == 3
                bz, seq_len, channels = r.shape
                assert channels % head_size == 0
                assert seq_len <= max_sequence_length
                assert reduce(lambda x, y: x * y, u.shape, 1) == channels, (
                    "the elements of u (time first) is not equal to hidden_size"
                )

                r_dtype = dtypes.canonicalize_dtype(r.dtype)
                k_dtype = dtypes.canonicalize_dtype(k.dtype)
                v_dtype = dtypes.canonicalize_dtype(v.dtype)
                w_dtype = dtypes.canonicalize_dtype(w.dtype)
                u_dtype = dtypes.canonicalize_dtype(u.dtype)
                gy_dtype = dtypes.canonicalize_dtype(gy.dtype)

                u_shape = (bz,) + u_shape
                assert all(
                    [r_dtype == xx for xx in [k_dtype, v_dtype, w_dtype, u_dtype]]
                )
                assert r_dtype in [jnp.float32, jnp.float16, jnp.bfloat16]
                if r_dtype in [jnp.float32, jnp.bfloat16]:
                    output_dtype = r_dtype
                else:
                    output_dtype = jnp.float32
                assert output_dtype == gy_dtype

                outputs = [
                    ShapedArray(r.shape, output_dtype, named_shape=r.named_shape),
                    ShapedArray(k.shape, output_dtype, named_shape=k.named_shape),
                    ShapedArray(v.shape, output_dtype, named_shape=v.named_shape),
                    ShapedArray(w.shape, output_dtype, named_shape=w.named_shape),
                    ShapedArray(u_shape, output_dtype, named_shape=u.named_shape),
                ]  # output
                return outputs

            _rwkv_bwd_p.def_abstract_eval(_rwkv_bwd_abstract)
            return _rwkv_bwd, _rwkv_bwd_p

        rwkv_bwd, _ = build_backward_op()
        """
        组合算子
        """

        @jax.custom_vjp
        def rwkv_normal_operator(r, k, v, w, u):
            output, _ = rwkv_fwd(r, k, v, w, u)
            return output

        rwkv_normal_operator.defvjp(rwkv_fwd, rwkv_bwd)
        self.rwkv_normal_operator = rwkv_normal_operator

        def build_forward_with_state_op():
            _rwkv_fwd_state_p = core.Primitive("rwkv_fwd_with_state")
            _rwkv_fwd_state_p.multiple_results = True
            _rwkv_fwd_state_p.def_impl(partial(xla.apply_primitive, _rwkv_fwd_state_p))

            """
            构建带有状态前向过程方法
            """

            def _rwkv_fwd_with_state(r, k, v, w, u, init_state=None, state_map=None):
                bz = r.shape[0]
                if init_state is not None:  # shape=(B,H,D,D)
                    assert len(init_state.shape) in [
                        3,
                        4,
                    ], "init_state的shape为(Batch_size,num_heads,head_size,head_size)"
                    if len(init_state.shape) == 3:
                        state_map = jnp.zeros((bz,), dtype=jnp.int32)

                    if state_map is None:
                        n_state = jnp.shape(init_state)[0]
                        B = jnp.shape(r)[0]
                        # print('ns:',n_state,'B:',B,r.shape,k.shape,v.shape)
                        assert n_state == 1 or n_state == B, (
                            "我无法为您推断state_map的形状，请手动指定。"
                        )
                        if n_state == 1:
                            state_map = jnp.array([0] * B, dtype=jnp.int32)
                        elif n_state == B:
                            state_map = jnp.array(
                                [i for i in range(B)], dtype=jnp.int32
                            )
                        else:
                            assert False, "未实现"
                    else:
                        # assert state_map is not None,"请传入一个state_map,这是一个int32类型的shape为(bz,)的数组，存放的是int_state到每一维度上的映射关系"
                        assert state_map.dtype in [
                            jnp.int64,
                            jnp.int32,
                        ], "state_map的数值类型必须为int32"
                        state_map = jnp.astype(state_map, jnp.int32)
                        assert jnp.all(state_map >= 0) and jnp.add(state_map < bz), (
                            f"state_map内为state的映射下标，因此范围为: [0,{bz})"
                        )
                assert (init_state is None) == (state_map is None), (
                    "init_state与state_map必须同时传入"
                )

                if init_state is None:
                    y, s = _rwkv_fwd_state_p.bind(r, k, v, w, u)
                else:
                    # state_map = jnp.astype(state_map,jnp.int32)
                    # print('sd',state_map.dtype)
                    y, s = _rwkv_fwd_state_p.bind(r, k, v, w, u, init_state, state_map)
                return y, s

            """
            milr调用上面注册的前向过程算子，milr帮助算子创建输出张量
            """

            def _rwkv_fwd_with_state_cuda_lowering(
                ctx, r, k, v, w, u, s=None, s_map=None
            ):
                r_type = ir.RankedTensorType(r.type)
                k_type = ir.RankedTensorType(k.type)
                v_type = ir.RankedTensorType(v.type)
                w_type = ir.RankedTensorType(w.type)
                u_type = ir.RankedTensorType(u.type)

                with_init_state = s is not None

                assert all(
                    [
                        r_type.element_type == xx.element_type
                        for xx in [k_type, v_type, w_type, u_type]
                    ]
                )
                assert all(
                    [r_type.shape == xx.shape for xx in [k_type, v_type, w_type]]
                )
                assert r_type.element_type in [
                    ir.F32Type.get(),
                    ir.BF16Type.get(),
                    ir.F16Type.get(),
                ]
                bz, seq_len, hd_sz = r_type.shape

                assert hd_sz % head_size == 0
                num_heads = hd_sz // head_size
                assert reduce(lambda x, y: x * y, u_type.shape, 1) == hd_sz, (
                    "the elements of u (time first) is not equal to hidden_size"
                )
                input_type = r_type.element_type

                if input_type in [ir.F32Type.get(), ir.BF16Type.get()]:
                    output_type = input_type
                else:
                    output_type = ir.F32Type.get()
                state_shape = (bz, num_heads, head_size, head_size)

                if with_init_state:
                    assert s_map is not None, (
                        "您必须同时传入init_state与state_map 或者都赋值为None."
                    )

                    s_type = ir.RankedTensorType(s.type)
                    sm_type = ir.RankedTensorType(s_map.type)
                    # print(sm_type, ir.IntegerType.get_signless(64))
                    assert sm_type.element_type == ir.IntegerType.get_signless(32), (
                        "state_map的数据类型必须为int32"
                    )
                    # print(sm_type.shape,bz)
                    assert tuple(sm_type.shape) == (bz,), (
                        "state_map的shape 形状必须为(batch_size,)"
                    )

                    assert s_type.element_type == output_type
                    assert tuple(s_type.shape) == state_shape, (
                        "the shape of init state must be (batch_size,num_heads,head_size,head_size)"
                    )
                    # assert s_type.shape[0] == bz and reduce(lambda x,y: x * y, s_type.shape[1:],1) == head_size * hd_sz,"the shape of init state must be (batch_size,num_heads,head_size,head_size)"

                opaque = rwkv_kernel.create_rwkv_descriptor(
                    bz,
                    seq_len,
                    hd_sz,
                    hd_sz // head_size,
                    with_init_state,
                    element_type_to_descriptor_type_mapping(input_type),
                    element_type_to_descriptor_type_mapping(output_type),
                )

                if with_init_state:
                    operands = [s_map, r, k, v, w, u, s]
                    operand_layouts = default_layouts(
                        sm_type.shape,
                        r_type.shape,
                        k_type.shape,
                        v_type.shape,
                        w_type.shape,
                        u_type.shape,
                        s_type.shape,
                    )
                else:
                    operands = [r, k, v, w, u]
                    operand_layouts = default_layouts(
                        r_type.shape,
                        k_type.shape,
                        v_type.shape,
                        w_type.shape,
                        u_type.shape,
                    )

                out = custom_call(
                    b"wkv_forward_with_state",
                    result_types=[
                        ir.RankedTensorType.get(r_type.shape, output_type),
                        ir.RankedTensorType.get(state_shape, output_type),
                    ],
                    operands=operands,
                    backend_config=opaque,
                    operand_layouts=operand_layouts,
                    result_layouts=default_layouts(r_type.shape, state_shape),
                ).results
                return out

            """
            将算子绑定到C++
            """
            mlir.register_lowering(
                _rwkv_fwd_state_p,
                _rwkv_fwd_with_state_cuda_lowering,
                platform="gpu",
            )
            """
            定义抽象过程，告知jax输出张量形状与数据类型
            """

            def _rwkv_fwd_with_state_abstract(r, k, v, w, u, s=None, s_map=None):
                assert all([r.shape == xx.shape for xx in [k, v, w]])
                assert len(r.shape) == 3
                bz, seq_len, channels = r.shape
                assert channels % head_size == 0
                assert seq_len <= max_sequence_length
                assert reduce(lambda x, y: x * y, u.shape, 1) == channels, (
                    "the elements of u (time first) is not equal to hidden_size"
                )
                num_heads = channels // head_size
                r_dtype = dtypes.canonicalize_dtype(r.dtype)
                k_dtype = dtypes.canonicalize_dtype(k.dtype)
                v_dtype = dtypes.canonicalize_dtype(v.dtype)
                w_dtype = dtypes.canonicalize_dtype(w.dtype)
                u_dtype = dtypes.canonicalize_dtype(u.dtype)

                assert all(
                    [r_dtype == xx for xx in [k_dtype, v_dtype, w_dtype, u_dtype]]
                )
                assert r_dtype in [jnp.float32, jnp.float16, jnp.bfloat16]
                if r_dtype in [jnp.float32, jnp.bfloat16]:
                    output_dtype = r_dtype
                else:
                    output_dtype = jnp.float32

                state_shape = (bz, num_heads, head_size, head_size)
                if s is not None:
                    s_dtype = dtypes.canonicalize_dtype(s.dtype)
                    assert s_dtype == output_dtype
                    assert s.shape == state_shape, (
                        "the shape of init_state must be (batch_size, seq_len, num_heads, head_size, head_size)"
                    )

                return [
                    ShapedArray(
                        r.shape, output_dtype, named_shape=r.named_shape
                    ),  # output
                    ShapedArray(state_shape, output_dtype, named_shape=r.named_shape),
                ]

            _rwkv_fwd_state_p.def_abstract_eval(_rwkv_fwd_with_state_abstract)
            return _rwkv_fwd_with_state, _rwkv_fwd_state_p

        self.rwkv_fwd_with_state, _ = build_forward_with_state_op()

    def __call__(
        self, r, k, v, w, u, with_state=False, init_state=None, state_map=None
    ):
        if with_state:
            return self.rwkv_fwd_with_state(r, k, v, w, u, init_state, state_map)
        else:
            return self.rwkv_normal_operator(r, k, v, w, u), None

    @staticmethod
    def _load_or_build_kernel(head_size, max_sequence_length):
        assert head_size % 4 == 0, f"head size必须是4的倍数，而{head_size}显然不是."
        assert isinstance(head_size, int), "你是在搞笑吗？ head_size肯定得是int类型的啊"
        assert isinstance(max_sequence_length, int), (
            "你是在搞笑吗？ max_sequence_length肯定得是int类型的啊"
        )
        assert head_size > 0 and max_sequence_length > 0, (
            "难绷，head_size与max_sequence_length肯定得是大于0的正整数啊。"
        )
        assert os.path.exists(cuda_lib_dir) and len(os.listdir(cuda_lib_dir)) > 0, (
            f"请检查{cuda_lib_dir}文件夹是否存在，这个文件本质是是您的cuda library的超链接。"
        )
        kernel_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), kernel_dir_name)
        )
        builds_dir = os.path.join(kernel_dir, "builds")
        assert os.path.exists(kernel_dir), (
            f"找不到{kernel_dir_name}文件夹，请问您的文件是完整的吗？"
        )
        if not os.path.exists(builds_dir):
            os.mkdir(builds_dir)
        target_dir_name = f"_N_{head_size}_T_{max_sequence_length}"
        target_dir = os.path.join(builds_dir, target_dir_name)
        if not os.path.exists(target_dir):
            os.mkdir(target_dir)

        def get_cflags():
            getvar = sysconfig.get_config_var
            flags = [
                "-I" + sysconfig.get_path("include"),
                "-I" + sysconfig.get_path("platinclude"),
            ]

            flags.extend(getvar("CFLAGS").split())
            return " ".join(flags)

        def get_suffix():
            getvar = sysconfig.get_config_var
            return getvar("EXT_SUFFIX")

        build_cmds = []

        # first, build cuda kernel
        if use_rocm:
            cu_src = os.path.join(kernel_dir, "rwkv_kernels.hip")
            assert os.path.exists(cu_src)
            cu_dst = os.path.join(target_dir, "rwkv_kernels.hip.o")
            kernel_cmd = (
                "hipcc -O3 --hipstdpar -xhip -fopenmp -ffast-math"
                + " -munsafe-fp-atomics -enable-vectorize-compares"
                + f" -I{cuda_lib_dir} -I{pybind11.get_include()}"
                + " -fPIC -D__HIP_PLATFORM_AMD__=1 -DUSE_ROCM=1 -DHIPBLAS_V2"
                + " --gpu-max-threads-per-block=120"
                + f" -c {cu_src} -o {cu_dst} -D _N_={head_size} -D _T_={max_sequence_length}"
            )
        else:
            cu_src = os.path.join(kernel_dir, "rwkv_kernels.cu")
            assert os.path.exists(cu_src)
            cu_dst = os.path.join(target_dir, "rwkv_kernels.cu.o")
            kernel_cmd = (
                "nvcc --threads 4 -Xcompiler -Wall -ldl --expt-relaxed-constexpr -O3 -DNDEBUG -Xcompiler -O3"
                + " --generate-code=arch=compute_70,code=[compute_70,sm_70] --generate-code=arch=compute_75,code=[compute_75,sm_75] --generate-code=arch=compute_80,code=[compute_80,sm_80] --generate-code=arch=compute_86,code=[compute_86,sm_86]"
                + f" -Xcompiler=-fPIC -Xcompiler=-fvisibility=hidden -x cu -c {cu_src} -o {cu_dst} -D _N_={head_size} -D _T_={max_sequence_length}"
            )
        build_cmds.append(kernel_cmd)

        so_dst = os.path.join(target_dir, f"{kernel_name}{get_suffix()}")
        if not os.path.exists(so_dst):
            # second, build C++ code.
            cpp_src = os.path.join(kernel_dir, f"{kernel_name}.cpp")
            cpp_dst = os.path.join(builds_dir, f"{kernel_name}.cpp.o")
            if not os.path.exists(cpp_dst):
                if use_rocm:
                    cpp_cmd = (
                        f"c++ -I{cuda_lib_dir} -I{pybind11.get_include()} {get_cflags()}"
                        + " -fPIC -D__HIP_PLATFORM_AMD__=1 -DUSE_ROCM=1 -DHIPBLAS_V2"
                        + " -O3 -DNDEBUG -O3 -fPIC -fvisibility=hidden -flto -fno-fat-lto-objects"
                        + f" -o {cpp_dst} -c {cpp_src}"
                    )
                else:
                    cpp_cmd = (
                        f"c++ -I{cuda_lib_dir} -I{pybind11.get_include()} {get_cflags()}"
                        + " -O3 -DNDEBUG -O3 -fPIC -fvisibility=hidden -flto -fno-fat-lto-objects"
                        + f" -o {cpp_dst} -c {cpp_src}"
                    )
                build_cmds.append(cpp_cmd)

            # third assembly C++ and cuda
            if use_rocm:
                assembly_cmd = (
                    f"c++ -fPIC -O3 -DNDEBUG -O3 -flto -shared  -o {so_dst} {cpp_dst} {cu_dst}"
                    + f" -fPIC -I{cuda_lib_dir} -I{pybind11.get_include()} {get_cflags()}"
                    + " -D__HIP_PLATFORM_AMD__=1 -DUSE_ROCM=1 -DHIPBLAS_V2"
                    + " -L/opt/rocm/lib  -lamdhip64 -lpthread -ldl"
                )
            else:
                assembly_cmd = (
                    f"c++ -fPIC -O3 -DNDEBUG -O3 -flto -shared  -o {so_dst} {cpp_dst} {cu_dst}"
                    + " -L/usr/local/cuda/lib64  -lcudadevrt -lcudart_static -lrt -lpthread -ldl"
                )
            build_cmds.append(assembly_cmd)

            # finally strip the so library
            strip_cmd = f"strip {so_dst}"
            build_cmds.append(strip_cmd)

            print("-------------------starting build kernel -------------------")
            for cmd in build_cmds:
                print("--------------- execute cmd ---------------")
                print(cmd)
                p = subprocess.Popen(cmd, shell=True)
                p.wait()
            print("-------------------build kernel finished -------------------")
        kernel_module_path = f"rwkv_ops.rwkv6_kernel.{kernel_dir_name}.builds.{target_dir_name}.{kernel_name}"
        print("loading cuda kernel from the path:", kernel_module_path)
        rwkv_op = importlib.import_module(kernel_module_path)
        return rwkv_op


if __name__ == "__main__":
    bz, seq_len, hd_sz = 1, 16, 8
    r = jnp.zeros(shape=(bz, seq_len, hd_sz), dtype="float16") + 2
    k = jnp.zeros(shape=(bz, seq_len, hd_sz), dtype="float16") + 2
    v = jnp.zeros(shape=(bz, seq_len, hd_sz), dtype="float16") + 2
    w = jnp.zeros(shape=(bz, seq_len, hd_sz), dtype="float16") - 2
    u = jnp.zeros(shape=(hd_sz,), dtype="float16") + 2
    rwkv_op = RWKVKernelOperator(head_size=hd_sz, max_sequence_length=seq_len)
    out = rwkv_op(r, k, v, w, u)

    print(out.dtype)

    def ref_loss(r, k, v, w, u):
        predictions = rwkv_op(r, k, v, w, u)
        return -jnp.mean(predictions**2)

    ref_out = jax.grad(ref_loss, argnums=(0, 1, 2, 3, 4))(r, k, v, w, u)
    print(ref_out)



import os
import torch
from torch.utils.cpp_extension import load

kernel_dir_name = "torch_kernel"

use_rocm = "RWKV_USE_ROCM" in os.environ and os.environ["RWKV_USE_ROCM"] == "1"


class RWKVKernelOperator:
    def __init__(self, head_size, max_sequence_length):
        current_dir = os.path.dirname(__file__)
        # current_dir = os.pat
        if use_rocm:
            wkv6_cuda = load(
                name="wkv6",
                sources=[
                    os.path.join(current_dir, f"{kernel_dir_name}/wkv6_op.cpp"),
                    os.path.join(current_dir, f"{kernel_dir_name}/wkv6_cuda.cu"),
                ],
                # verbose=True, extra_cuda_cflags=[f"-D_N_={head_size}", f"-D_T_={max_sequence_length}"])
                verbose=True,
                extra_cuda_cflags=[
                    "-fopenmp -ffast-math -munsafe-fp-atomics --gpu-max-threads-per-block=120 -enable-vectorize-compares",
                    f"-D_N_={head_size}",
                    f"-D_T_={max_sequence_length}",
                ],
            )
        else:
            wkv6_cuda = load(
                name="wkv6",
                sources=[
                    os.path.join(current_dir, f"{kernel_dir_name}/wkv6_op.cpp"),
                    os.path.join(current_dir, f"{kernel_dir_name}/wkv6_cuda.cu"),
                ],
                # verbose=True, extra_cuda_cflags=[f"-D_N_={head_size}", f"-D_T_={max_sequence_length}"])
                verbose=True,
                extra_cuda_cflags=[
                    "-res-usage",
                    "--use_fast_math",
                    "-O3",
                    "-Xptxas -O3",
                    "--extra-device-vectorization",
                    f"-D_N_={head_size}",
                    f"-D_T_={max_sequence_length}",
                ],
            )

        class RWKV_6(torch.autograd.Function):
            @staticmethod
            def forward(ctx, B, T, C, H, r, k, v, w, u):
                if not isinstance(u, torch.Tensor):
                    u = u.value
                with torch.no_grad():
                    assert r.dtype == k.dtype == v.dtype == w.dtype == u.dtype
                    assert r.dtype in [torch.float32, torch.bfloat16, torch.float16]

                    assert head_size == C // H
                    ctx.B = B
                    ctx.T = T
                    ctx.C = C
                    ctx.H = H
                    assert r.is_contiguous()
                    assert k.is_contiguous()
                    assert v.is_contiguous()
                    assert w.is_contiguous()
                    assert u.is_contiguous()
                    ctx.save_for_backward(r, k, v, w, u)

                    y_dtype = r.dtype if r.dtype != torch.float16 else torch.float32

                    y = torch.empty(
                        (B, T, C),
                        device=r.device,
                        dtype=y_dtype,
                        memory_format=torch.contiguous_format,
                    )  # .uniform_(-100, 100)

                    if r.dtype == torch.float32:
                        wkv6_cuda.forward_fp32(B, T, C, H, r, k, v, w, u, y)
                    elif r.dtype == torch.bfloat16:
                        wkv6_cuda.forward_bf16(B, T, C, H, r, k, v, w, u, y)
                    else:
                        wkv6_cuda.forward_fp16(B, T, C, H, r, k, v, w, u, y)
                    return y

            @staticmethod
            def backward(ctx, gy):
                assert gy.is_cuda
                with torch.no_grad():
                    assert gy.dtype in [torch.bfloat16, torch.float32]
                    B = ctx.B
                    T = ctx.T
                    C = ctx.C
                    H = ctx.H
                    assert gy.is_contiguous()
                    r, k, v, w, u = ctx.saved_tensors
                    y_dtype = r.dtype if r.dtype != torch.float16 else torch.float32

                    gr = torch.empty(
                        (B, T, C),
                        device=gy.device,
                        requires_grad=False,
                        dtype=y_dtype,
                        memory_format=torch.contiguous_format,
                    )  # .uniform_(-100, 100)
                    gk = torch.empty(
                        (B, T, C),
                        device=gy.device,
                        requires_grad=False,
                        dtype=y_dtype,
                        memory_format=torch.contiguous_format,
                    )  # .uniform_(-100, 100)
                    gv = torch.empty(
                        (B, T, C),
                        device=gy.device,
                        requires_grad=False,
                        dtype=y_dtype,
                        memory_format=torch.contiguous_format,
                    )  # .uniform_(-100, 100)
                    gw = torch.empty(
                        (B, T, C),
                        device=gy.device,
                        requires_grad=False,
                        dtype=y_dtype,
                        memory_format=torch.contiguous_format,
                    )  # .uniform_(-100, 100)
                    gu = torch.empty(
                        (B, C),
                        device=gy.device,
                        requires_grad=False,
                        dtype=y_dtype,
                        memory_format=torch.contiguous_format,
                    )  # .uniform_(-100, 100)

                    if r.dtype == torch.float32:
                        wkv6_cuda.backward_fp32(
                            B, T, C, H, r, k, v, w, u, gy, gr, gk, gv, gw, gu
                        )
                    elif r.dtype == torch.bfloat16:
                        wkv6_cuda.backward_bf16(
                            B, T, C, H, r, k, v, w, u, gy, gr, gk, gv, gw, gu
                        )
                    else:
                        wkv6_cuda.backward_fp16(
                            B, T, C, H, r, k, v, w, u, gy, gr, gk, gv, gw, gu
                        )

                    gu = torch.sum(gu, 0).view(H, C // H)

                    return (None, None, None, None, gr, gk, gv, gw, gu)

        class RWKV_6_with_state:
            @staticmethod
            def apply(B, T, C, H, S, s_map, r, k, v, w, u, s):
                with torch.no_grad():
                    assert s_map.dtype == torch.int64, (
                        "s_map 必须为None 或者是长度为B的，int64类型的数组。"
                    )
                    assert (s is None and s_map is None) or (
                        s is not None and s_map is not None
                    ), "init_state与s_map必须同时为None 或者同时不为None"
                    assert (
                        r.dtype == k.dtype == v.dtype == w.dtype == u.dtype
                        and r.dtype in [torch.float16, torch.float32, torch.bfloat16]
                    ), " r, k, v, w, u 必须为fp16 fp32 bf16中的一种 并且类型相同"
                    if r.dtype in [torch.float32, torch.bfloat16]:
                        o_dtype = r.dtype
                    else:
                        o_dtype = torch.float32
                    assert (
                        r.device
                        == k.device
                        == v.device
                        == w.device
                        == u.device
                        == s.device
                        == s_map.device
                    ), "what kan i say? 请确保r k v w u s s_map在同一设备上，快去检查！"

                    y = torch.empty(
                        (B, T, C),
                        device=r.device,
                        dtype=o_dtype,
                        memory_format=torch.contiguous_format,
                    )
                    ys = torch.empty(
                        (B, H, head_size, head_size),
                        device=r.device,
                        dtype=o_dtype,
                        memory_format=torch.contiguous_format,
                    )
                    # print(ys)
                    if r.dtype == torch.bfloat16:
                        wkv6_cuda.forward_with_state_bf16(
                            B, T, C, H, S, s_map, r, k, v, w, u, s, y, ys
                        )
                    elif r.dtype == torch.float32:
                        wkv6_cuda.forward_with_state_fp32(
                            B, T, C, H, S, s_map, r, k, v, w, u, s, y, ys
                        )
                    else:
                        wkv6_cuda.forward_with_state_fp16(
                            B, T, C, H, S, s_map, r, k, v, w, u, s, y, ys
                        )

                return y, ys

        self.head_size = head_size
        self.normal_kernenl = RWKV_6
        self.kernel_with_state = RWKV_6_with_state

    def __call__(
        self, r, k, v, w, u, with_state=False, init_state=None, state_map=None
    ):
        B, T, C = r.shape
        assert C % self.head_size == 0
        H = C // self.head_size
        if not isinstance(u, torch.Tensor):
            u = u.value

        assert r.is_cuda
        assert k.is_cuda
        assert v.is_cuda
        assert w.is_cuda
        assert u.is_cuda

        if isinstance(r, torch.Tensor):
            assert r.device == k.device == v.device == w.device == u.device
        else:
            r.get_device() == k.get_device() == v.get_device() == w.get_device() == u.get_device()

        assert r.dtype == k.dtype == v.dtype == w.dtype == u.dtype

        if r.dtype in [torch.float32, torch.bfloat16]:
            s_dtype = r.dtype
        else:
            s_dtype = torch.float32

        is_custom_init = init_state is not None

        if init_state is not None:
            assert len(init_state.shape) in [3, 4], (
                "init_state 的形状必须为(state_kinds /*<= Batch_size*/,num_heads,head_size,head_size) 或者(num_heads,head_size,head_size)"
            )
            if len(init_state.shape) == 3:
                init_state = init_state[None, :]
            assert (
                init_state.shape[1:] == (H, self.head_size, self.head_size)
                and init_state.shape[0] <= B
            ), (
                "init_state 的形状必须为(state_kinds /*<= Batch_size*/,num_heads,head_size,head_size) 或者(num_heads,head_size,head_size)"
            )

            assert init_state.dtype == s_dtype, f"init_state的数值类型应为: {s_dtype}"
            assert init_state.device == r.device

        if state_map is not None:
            if isinstance(state_map, list):
                state_map = torch.tensor(state_map, dtype=torch.int64)
            elif isinstance(state_map, torch.Tensor):
                assert state_map.dtype in [
                    torch.int32,
                    torch.int64,
                ], "state_map是一个长度为Batch_Size的int64类型的映射数组"
                state_map = state_map.to(torch.int64)
            assert state_map.shape == (B,), "state_map的shape必须为(Batch_Size,)"
            assert state_map.device == r.deivec

        if with_state:
            if init_state is None:
                assert state_map is None, (
                    "您必须在指定了init_state的情况下才能使用state_map"
                )
                init_state = torch.zeros((0,), device=r.device, dtype=s_dtype)
                state_map = torch.zeros((0,), device=r.device, dtype=torch.int64)
            else:
                n_state = init_state.shape[0]
                if state_map is None:
                    assert n_state == 1 or n_state == B, (
                        "我无法为您推断state_map的形状，请手动指定。"
                    )
                    if n_state == 1:
                        state_map = torch.tensor(
                            [0] * B, dtype=torch.int64, device=r.device
                        )
                    elif n_state == B:
                        state_map = torch.tensor(
                            [i for i in range(B)], dtype=torch.int64, device=r.device
                        )
                    else:
                        assert False, "未实现"
                else:
                    assert state_map.shape == (B,), "state_map的形状必须为(batch_size,)"
                    assert (state_map >= 0).all() and (state_map < n_state).all(), (
                        f"state_map的取值范围为[0,{n_state})之间的整数，您的输入显然不满足。"
                    )
            # print('state map:',state_map)
            o, ys = self.kernel_with_state.apply(
                B, T, C, H, is_custom_init, state_map, r, k, v, w, u, init_state
            )
            return o, ys
        else:
            o = self.normal_kernenl.apply(B, T, C, H, r, k, v, w, u)
            return o, None



from ..triton_kernel.mhc_pre_op import *
import jax
import jax.numpy as jnp
import jax_triton as jt
from functools import partial
import jax.tree_util as jtu

# 引入自定义分区器和切分规则 (适配 JAX 新版 Shardy 引擎)
from jax.experimental.custom_partitioning import custom_partitioning

# =========================================================================
# 【核心配置】：为 Shardy 引擎定义的静态 Einsum 切分映射字符串
# 字母含义: b=Batch, t=Time, n=NSize1, m=NSize2, c=Channel
# =========================================================================
BATCH_AXIS_NAME = "data"

# FWD_RULE 含义:
# 输入1(x):        b t n c
# 输入2(h_res_in): b t n m
# 输入3(h_pre_in): b t n
# 输出1(out):      b t c
# 输出2(H_res_out):b t n m
FWD_RULE = "b t n c, b t n m, b t n -> b t c, b t n m"

# BWD_RULE 含义:
# 输入1-3同上
# 输入4(g_out):    b t c
# 输入5(g_H_res):  b t n m
# 输出1(gx):       b t n c
# 输出2(gh_res):   b t n m
# 输出3(gh_pre):   b t n
BWD_RULE = "b t n c, b t n m, b t n, b t c, b t n m -> b t n c, b t n m, b t n"


# 兼容老版本的推导函数 (接收末尾传来的静态参数 *static_args)
def _fwd_infer_sharding(arg_shapes, arg_shardings, *static_args):
    qs = arg_shardings[0]
    return (qs, qs)


def _bwd_infer_sharding(arg_shapes, arg_shardings, *static_args):
    qs = arg_shardings[0]
    return (qs, qs, qs)


# JAX 官方标准的 partition 回调生成器 (接收末尾传来的静态参数 *static_args)
def _create_partition(impl_fn):
    def partition(mesh, arg_shapes, result_shape, *static_args):
        def lower_fn(*args):
            # 将动态 args 和 静态 static_args 拼在一起传给底层
            return impl_fn(*args, *static_args)

        result_shardings = jtu.tree_map(lambda x: x.sharding, result_shape)
        arg_shardings = jtu.tree_map(lambda x: x.sharding, arg_shapes)
        return mesh, lower_fn, result_shardings, arg_shardings

    return partition


# =========================================================================


# --- 1. 前向 Launcher ---
def mhc_pre_op_fwd_kernel_call(x, h_res_in, h_pre_in, num_iters, eps):
    B, T, n, C = x.shape
    total_bt = B * T

    # 重塑视图
    x_v = x.reshape(total_bt, n, C)
    h_res_v = h_res_in.reshape(total_bt, n, n)
    h_pre_v = h_pre_in.reshape(total_bt, n)

    # 计算输入步长
    sx_bt, sx_n, sx_c = jt.strides_from_shape(x_v.shape)
    shr_bt, shr_n1, shr_n2 = jt.strides_from_shape(h_res_v.shape)
    shp_bt, shp_n = jt.strides_from_shape(h_pre_v.shape)

    # 准备输出结构 (注意：out 是 BF16/FP16等, H_res_out 是 FP32)
    out_shapes = [
        jax.ShapeDtypeStruct((total_bt, C), x.dtype),  # out_ptr
        jax.ShapeDtypeStruct((total_bt, n, n), jnp.float32),  # H_res_out_ptr
    ]

    # 计算输出步长
    so_bt, so_c = jt.strides_from_shape(out_shapes[0].shape)
    sHr_bt, sHr_n1, sHr_n2 = jt.strides_from_shape(out_shapes[1].shape)

    # 定义 Grid
    grid = lambda meta: (
        jt.cdiv(total_bt, meta["BLOCK_BT"]),
        jt.cdiv(C, meta["BLOCK_C"]),
    )

    # 调用 Triton
    out_v, H_res_out_v = jt.triton_call(
        x_v,
        h_res_v,
        h_pre_v,  # 指针参数 1, 2, 3
        kernel=sinkhorn_aggregate_fused_kernel,
        out_shape=out_shapes,  # 产生指针参数 4, 5
        grid=grid,
        # --- 标量参数 ---
        Total_BT_CONST=total_bt,
        NSIZE=n,
        CSIZE=C,
        NUM_ITERS=num_iters,
        EPS=eps,
        stride_x_bt=sx_bt,
        stride_x_n=sx_n,
        stride_x_c=sx_c,
        stride_h_res_in_bt=shr_bt,
        stride_h_res_in_n1=shr_n1,
        stride_h_res_in_n2=shr_n2,
        stride_h_pre_in_bt=shp_bt,
        stride_h_pre_in_n=shp_n,
        stride_out_bt=so_bt,
        stride_out_c=so_c,
        stride_Hr_out_bt=sHr_bt,
        stride_Hr_out_n1=sHr_n1,
        stride_Hr_out_n2=sHr_n2,
    )

    return out_v.reshape(B, T, C), H_res_out_v.reshape(B, T, n, n)


# =========================================================================
# 【新增】：包装前向算子 (适配 SPMD 并行，修复调用方式)
# =========================================================================
def _mhc_pre_op_fwd_spmd_impl(x, h_res_in, h_pre_in, num_iters, eps):
    return mhc_pre_op_fwd_kernel_call(x, h_res_in, h_pre_in, num_iters, eps)


mhc_pre_op_fwd_spmd = custom_partitioning(
    _mhc_pre_op_fwd_spmd_impl, static_argnums=(3, 4)
)

mhc_pre_op_fwd_spmd.def_partition(
    infer_sharding_from_operands=_fwd_infer_sharding,
    sharding_rule=FWD_RULE,
    partition=_create_partition(mhc_pre_op_fwd_kernel_call),
)


# --- 2. 反向 Launcher ---
def mhc_pre_op_bwd_kernel_call(x, h_res, h_pre, grad_out, grad_H_res, num_iters, eps):
    B, T, n, C = x.shape
    total_bt = B * T

    # 视图重塑
    x_v = x.reshape(total_bt, n, C)
    h_v = h_res.reshape(total_bt, n, n)
    h_p_v = h_pre.reshape(total_bt, n)
    g_out_v = grad_out.reshape(total_bt, C)
    g_H_v = grad_H_res.reshape(total_bt, n, n)

    # 步幅
    sgout_bt, sgout_c = jt.strides_from_shape(g_out_v.shape)
    sgH_bt, sgH_n1, sgH_n2 = jt.strides_from_shape(g_H_v.shape)
    sx_bt, sx_n, sx_c = jt.strides_from_shape(x_v.shape)
    sh_bt, sh_n1, sh_n2 = jt.strides_from_shape(h_v.shape)
    shp_bt, shp_n = jt.strides_from_shape(h_p_v.shape)

    # 准备输出梯度结构 (gx, gh_res, gh_pre)
    out_shapes = [
        jax.ShapeDtypeStruct(x_v.shape, x.dtype),
        jax.ShapeDtypeStruct(h_v.shape, h_res.dtype),
        jax.ShapeDtypeStruct(h_p_v.shape, h_pre.dtype),
    ]

    # 获取输出步幅传给反向 kernel
    sgx_bt, sgx_n, sgx_c = sx_bt, sx_n, sx_c
    sghr_bt, sghr_n1, sghr_n2 = sh_bt, sh_n1, sh_n2
    sghp_bt, sghp_n = shp_bt, shp_n

    # 调用 Triton
    gx_v, gh_res_v, gh_pre_v = jt.triton_call(
        g_out_v,
        g_H_v,
        x_v,
        h_v,
        h_p_v,  # 输入指针 1, 2, 3, 4, 5
        kernel=sinkhorn_aggregate_bwd_kernel,
        out_shape=out_shapes,  # 产生输出指针 6, 7, 8
        grid=(total_bt, 1),
        # --- 标量参数 ---
        TOTAL_BT_CONST=total_bt,
        NSIZE=n,
        CHANNEL_SIZE=C,
        NUM_ITERS=num_iters,
        EPS=eps,
        stride_gout_bt=sgout_bt,
        stride_gout_c=sgout_c,
        stride_gH_bt=sgH_bt,
        stride_gH_n1=sgH_n1,
        stride_gH_n2=sgH_n2,
        stride_x_bt=sx_bt,
        stride_x_n=sx_n,
        stride_x_c=sx_c,
        stride_h_res_bt=sh_bt,
        stride_h_res_n1=sh_n1,
        stride_h_res_n2=sh_n2,
        stride_h_pre_bt=shp_bt,
        stride_h_pre_n=shp_n,
        stride_gx_bt=sgx_bt,
        stride_gx_n=sgx_n,
        stride_gx_c=sgx_c,
        stride_gh_res_bt=sghr_bt,
        stride_gh_res_n1=sghr_n1,
        stride_gh_res_n2=sghr_n2,
        stride_gh_pre_bt=sghp_bt,
        stride_gh_pre_n=sghp_n,
    )

    return (
        gx_v.reshape(B, T, n, C),
        gh_res_v.reshape(B, T, n, n),
        gh_pre_v.reshape(B, T, n),
    )


# =========================================================================
# 【新增】：包装反向算子 (适配 SPMD 并行，修复调用方式)
# =========================================================================
def _mhc_pre_op_bwd_spmd_impl(x, h_res, h_pre, grad_out, grad_H_res, num_iters, eps):
    return mhc_pre_op_bwd_kernel_call(
        x, h_res, h_pre, grad_out, grad_H_res, num_iters, eps
    )


mhc_pre_op_bwd_spmd = custom_partitioning(
    _mhc_pre_op_bwd_spmd_impl, static_argnums=(5, 6)
)

mhc_pre_op_bwd_spmd.def_partition(
    infer_sharding_from_operands=_bwd_infer_sharding,
    sharding_rule=BWD_RULE,
    partition=_create_partition(mhc_pre_op_bwd_kernel_call),
)


# --- 3. JAX 接口绑定 (解决 Tracer Leak) ---


@partial(jax.jit, static_argnums=(3, 4))
def mhc_pre_op_fused(x, h_res_in, h_pre_in, num_iters=20, eps=1e-8):
    """
    通过闭包捕获静态参数 num_iters 和 eps，确保它们不进入 custom_vjp 的追踪范围。
    """

    @jax.custom_vjp
    def _internal_op(x_arr, hr_arr, hp_arr):
        # 【修改】：调用切分规则封装后的 FWD
        return mhc_pre_op_fwd_spmd(x_arr, hr_arr, hp_arr, num_iters, eps)

    def _internal_fwd(x_arr, hr_arr, hp_arr):
        out_tuple = _internal_op(x_arr, hr_arr, hp_arr)
        # 只保存参与微分的张量
        return out_tuple, (x_arr, hr_arr, hp_arr)

    def _internal_bwd(res, grads):
        x_arr, hr_arr, hp_arr = res
        grad_out, grad_H_res = grads

        # 处理可能的 None 梯度（虽然 JAX 通常会传零 Tensor，但为了健壮性）
        if grad_out is None:
            grad_out = jnp.zeros_like(x_arr[:, :, 0])
        if grad_H_res is None:
            grad_H_res = jnp.zeros_like(hr_arr)

        # 【修改】：调用切分规则封装后的 BWD
        gx, ghr, ghp = mhc_pre_op_bwd_spmd(
            x_arr, hr_arr, hp_arr, grad_out, grad_H_res, num_iters, eps
        )
        return gx, ghr, ghp

    _internal_op.defvjp(_internal_fwd, _internal_bwd)

    return _internal_op(x, h_res_in, h_pre_in)



"""
JAX 版 RWKV7 wkv kernel + generalized_delta_rule
延迟编译 CUDA 扩展，接口与 Torch 版本 1:1 对齐
"""

from __future__ import annotations
import pathlib
import subprocess
import ctypes
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from typing import Optional, Tuple, Union

# 引入自定义分区器 (适配 JAX 新版 Shardy 引擎)
from jax.experimental.custom_partitioning import custom_partitioning

CHUNK_LEN = 16  # 这是一个常数
# ---------- 延迟编译（改到当前目录） ----------
_CURRENT_DIR = pathlib.Path(
    __file__
).parent.absolute()  # rwkv_ops/rwkv7_kernel/jax_cuda_kernel

# =========================================================================
# 【核心修复】：为 Shardy 引擎定义的静态 Einsum 切分映射字符串
# 字母含义: b=Batch, t=Time, h=Head, k=HeadDim1, v=HeadDim2, c=Chunk
# =========================================================================
FWD_RULE = "b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b h k v -> b t h k, b h c k v, b t h k"
BWD_RULE = "b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b h c k v, b t h k, b h k v -> b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b h k v"
FWD_MASK_RULE = "b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b h k v, b t -> b t h k, b h c k v, b t h k"
BWD_MASK_RULE = "b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b t, b t h k, b h c k v, b t h k, b h k v -> b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b h k v"
INF_RULE = (
    "b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b h k v -> b t h k, b h k v"
)
INF_MASK_RULE = "b t h k, b t h k, b t h k, b t h k, b t h k, b t h k, b h k v, b t -> b t h k, b h k v"


# 保留给老版本 JAX 兼容使用的回调函数
def _fwd_infer_sharding(arg_shapes, arg_shardings):
    qs = arg_shardings[1]
    return (qs, qs, qs)


def _bwd_infer_sharding(arg_shapes, arg_shardings):
    qs = arg_shardings[1]
    return (qs, qs, qs, qs, qs, qs, qs)


def _inf_infer_sharding(arg_shapes, arg_shardings):
    qs = arg_shardings[1]
    return (qs, qs)


# =========================================================================
# 【关键修复】：根据 JAX 官方文档提供的标准 Partition 样板函数
# =========================================================================
def _create_partition(impl_fn):
    def partition(mesh, arg_shapes, result_shape):
        def lower_fn(*args):
            return impl_fn(*args)

        # 获取最终分配给该算子的 sharding 信息
        result_shardings = jtu.tree_map(lambda x: x.sharding, result_shape)
        arg_shardings = jtu.tree_map(lambda x: x.sharding, arg_shapes)
        return mesh, lower_fn, result_shardings, arg_shardings

    return partition


# =========================================================================


def get_jax_generalized_delta_rule(HEAD_SIZE=64):
    _BUILD_DIR = _CURRENT_DIR / f"build_{HEAD_SIZE}"
    _SO_PATH = _CURRENT_DIR / f"build_{HEAD_SIZE}/wkv7.so"

    def _ensure_compiled() -> pathlib.Path:
        if _SO_PATH.exists():
            return _SO_PATH

        print("[rwkv7_jax] First use – compiling CUDA kernel…")
        src_dir = _CURRENT_DIR
        build_dir = _BUILD_DIR
        build_dir.mkdir(exist_ok=True)

        xla_include_dir = jax.ffi.include_dir()
        if not xla_include_dir:
            raise RuntimeError("jax.ffi.include_dir() 返回空，请检查 JAX >= 0.4.31")

        cuda_flags = [
            "-ftz=true",
            "-prec-div=false",
            "-prec-sqrt=false",
            "--use_fast_math",
            "-O3",
            "-Xptxas=-O3",
            "-res-usage",
            "--extra-device-vectorization",
            "-D_C_=64",
            f"-D_C_={HEAD_SIZE}",
            f"-D_CHUNK_LEN_={CHUNK_LEN}",
        ]

        cmake_args = [
            "cmake",
            "-S",
            str(src_dir),
            "-B",
            str(build_dir),
            "-DCMAKE_BUILD_TYPE=Release",
            f"-DCMAKE_INSTALL_PREFIX={_CURRENT_DIR}",
            f"-DXLA_INCLUDE_DIR={xla_include_dir}",
            f"-DCMAKE_CUDA_FLAGS={' '.join(cuda_flags)}",
        ]
        subprocess.check_call(cmake_args)
        subprocess.check_call(["cmake", "--build", str(build_dir), "-j"])
        subprocess.check_call(["cmake", "--install", str(build_dir)])

        if not _SO_PATH.exists():
            raise RuntimeError("Compilation failed – wkv7.so not found.")

        print("[rwkv7_jax] Compilation finished – output at", _SO_PATH)
        return _SO_PATH

    _lib = ctypes.CDLL(_ensure_compiled())

    jax.ffi.register_ffi_target(
        "wkv7_fwd", jax.ffi.pycapsule(_lib.Wkv7Fwd), platform="CUDA"
    )
    jax.ffi.register_ffi_target(
        "wkv7_bwd", jax.ffi.pycapsule(_lib.Wkv7Bwd), platform="CUDA"
    )
    jax.ffi.register_ffi_target(
        "wkv7_inference", jax.ffi.pycapsule(_lib.Wkv7Inference), platform="CUDA"
    )

    jax.ffi.register_ffi_target(
        "wkv7_fwd_with_mask", jax.ffi.pycapsule(_lib.Wkv7FwdWithMask), platform="CUDA"
    )
    jax.ffi.register_ffi_target(
        "wkv7_bwd_with_mask", jax.ffi.pycapsule(_lib.Wkv7BwdWithMask), platform="CUDA"
    )
    jax.ffi.register_ffi_target(
        "wkv7_inference_with_mask",
        jax.ffi.pycapsule(_lib.Wkv7InferenceWithMask),
        platform="CUDA",
    )

    def _transpose_head(x: jnp.ndarray, head_first: bool) -> jnp.ndarray:
        x = jnp.asarray(x, dtype=jnp.bfloat16)
        if head_first:
            return jnp.transpose(x, (0, 2, 1, 3))
        return x

    # =========================================================================
    # 前向 + 反向 kernel (无 Mask)
    # =========================================================================
    def _wkv7_kernel_impl(w, q, k, v, a, b, h0):
        B, T, H, K = q.shape
        dtype = q.dtype
        chunk_num = int(T // CHUNK_LEN)
        out_type = jax.ShapeDtypeStruct((B, T, H, K), dtype)
        s_type = jax.ShapeDtypeStruct((B, H, chunk_num, K, K), jnp.float32)
        sa_type = jax.ShapeDtypeStruct((B, T, H, K), jnp.float32)

        return jax.ffi.ffi_call(
            "wkv7_fwd", (out_type, s_type, sa_type), vmap_method="broadcast_all"
        )(w, q, k, v, a, b, h0)

    @custom_partitioning
    def _wkv7_kernel(w, q, k, v, a, b, h0):
        return _wkv7_kernel_impl(w, q, k, v, a, b, h0)

    _wkv7_kernel.def_partition(
        infer_sharding_from_operands=_fwd_infer_sharding,
        sharding_rule=FWD_RULE,
        partition=_create_partition(_wkv7_kernel_impl),
    )

    @jax.custom_vjp
    def wk7_kernel(
        w: jnp.ndarray,
        q: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        h0: jnp.ndarray,
    ):
        y, s, sa = _wkv7_kernel(w, q, k, v, a, b, h0)
        finnal_state = s[:, :, -1]
        return (y, jnp.transpose(finnal_state, [0, 1, 3, 2]))

    def _fwd(
        w: jnp.ndarray,
        q: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        h0: jnp.ndarray,
    ):
        y, s, sa = _wkv7_kernel(w, q, k, v, a, b, h0)
        finnal_state = s[:, :, -1]
        return (y, jnp.transpose(finnal_state, [0, 1, 3, 2])), (w, q, k, v, a, b, s, sa)

    def _wkv7_bwd_kernel_impl(w, q, k, v, a, b, dy, s, sa, dht):
        dh0_type = jax.ShapeDtypeStruct(dht.shape, dht.dtype)
        dw_type = jax.ShapeDtypeStruct(w.shape, w.dtype)
        dq_type = jax.ShapeDtypeStruct(q.shape, q.dtype)
        dk_type = jax.ShapeDtypeStruct(k.shape, k.dtype)
        dv_type = jax.ShapeDtypeStruct(v.shape, v.dtype)
        da_type = jax.ShapeDtypeStruct(a.shape, a.dtype)
        db_type = jax.ShapeDtypeStruct(b.shape, b.dtype)

        dh0, dw, dq, dk, dv, da, db = jax.ffi.ffi_call(
            "wkv7_bwd",
            (dh0_type, dw_type, dq_type, dk_type, dv_type, da_type, db_type),
            vmap_method="broadcast_all",
        )(w, q, k, v, a, b, dy, s, sa, dht)
        return dw, dq, dk, dv, da, db, dh0

    @custom_partitioning
    def _wkv7_bwd_kernel(w, q, k, v, a, b, dy, s, sa, dht):
        return _wkv7_bwd_kernel_impl(w, q, k, v, a, b, dy, s, sa, dht)

    _wkv7_bwd_kernel.def_partition(
        infer_sharding_from_operands=_bwd_infer_sharding,
        sharding_rule=BWD_RULE,
        partition=_create_partition(_wkv7_bwd_kernel_impl),
    )

    def _bwd(res, grads):
        w, q, k, v, a, b, s, sa = res
        dy, dht = grads
        dy = jnp.asarray(dy, jnp.bfloat16)
        return _wkv7_bwd_kernel(w, q, k, v, a, b, dy, s, sa, dht)

    wk7_kernel.defvjp(_fwd, _bwd)

    # =========================================================================
    # 前向 + 反向 kernel (带 Mask)
    # =========================================================================
    def _wkv7_kernel_with_mask_impl(w, q, k, v, a, b, h0, mask):
        B, T, H, K = q.shape
        dtype = q.dtype
        chunk_num = int(T // CHUNK_LEN)
        out_type = jax.ShapeDtypeStruct((B, T, H, K), dtype)
        s_type = jax.ShapeDtypeStruct((B, H, chunk_num, K, K), jnp.float32)
        sa_type = jax.ShapeDtypeStruct((B, T, H, K), jnp.float32)

        return jax.ffi.ffi_call(
            "wkv7_fwd_with_mask",
            (out_type, s_type, sa_type),
            vmap_method="broadcast_all",
        )(w, q, k, v, a, b, mask, h0)

    @custom_partitioning
    def _wkv7_kernel_with_mask(w, q, k, v, a, b, h0, mask):
        return _wkv7_kernel_with_mask_impl(w, q, k, v, a, b, h0, mask)

    _wkv7_kernel_with_mask.def_partition(
        infer_sharding_from_operands=_fwd_infer_sharding,
        sharding_rule=FWD_MASK_RULE,
        partition=_create_partition(_wkv7_kernel_with_mask_impl),
    )

    @jax.custom_vjp
    def wk7_kernel_with_mask(
        w: jnp.ndarray,
        q: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        h0: jnp.ndarray,
        mask: jnp.ndarray,
    ):
        y, s, sa = _wkv7_kernel_with_mask(w, q, k, v, a, b, h0, mask)
        finnal_state = s[:, :, -1]
        return (y, jnp.transpose(finnal_state, [0, 1, 3, 2]))

    def _fwd_with_mask(
        w: jnp.ndarray,
        q: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        h0: jnp.ndarray,
        mask: jnp.ndarray,
    ):
        y, s, sa = _wkv7_kernel_with_mask(w, q, k, v, a, b, h0, mask)
        finnal_state = s[:, :, -1]
        return (y, jnp.transpose(finnal_state, [0, 1, 3, 2])), (
            w,
            q,
            k,
            v,
            a,
            b,
            s,
            sa,
            mask,
        )

    def _wkv7_bwd_kernel_with_mask_impl(w, q, k, v, a, b, mask, dy, s, sa, dht):
        dh0_type = jax.ShapeDtypeStruct(dht.shape, dht.dtype)
        dw_type = jax.ShapeDtypeStruct(w.shape, w.dtype)
        dq_type = jax.ShapeDtypeStruct(q.shape, q.dtype)
        dk_type = jax.ShapeDtypeStruct(k.shape, k.dtype)
        dv_type = jax.ShapeDtypeStruct(v.shape, v.dtype)
        da_type = jax.ShapeDtypeStruct(a.shape, a.dtype)
        db_type = jax.ShapeDtypeStruct(b.shape, b.dtype)

        dh0, dw, dq, dk, dv, da, db = jax.ffi.ffi_call(
            "wkv7_bwd_with_mask",
            (dh0_type, dw_type, dq_type, dk_type, dv_type, da_type, db_type),
            vmap_method="broadcast_all",
        )(w, q, k, v, a, b, mask, dy, s, sa, dht)
        return dw, dq, dk, dv, da, db, dh0

    @custom_partitioning
    def _wkv7_bwd_kernel_with_mask(w, q, k, v, a, b, mask, dy, s, sa, dht):
        return _wkv7_bwd_kernel_with_mask_impl(w, q, k, v, a, b, mask, dy, s, sa, dht)

    _wkv7_bwd_kernel_with_mask.def_partition(
        infer_sharding_from_operands=_bwd_infer_sharding,
        sharding_rule=BWD_MASK_RULE,
        partition=_create_partition(_wkv7_bwd_kernel_with_mask_impl),
    )

    def _bwd_with_mask(res, grads):
        w, q, k, v, a, b, s, sa, mask = res
        dy, dht = grads
        dy = jnp.asarray(dy, jnp.bfloat16)
        dw, dq, dk, dv, da, db, dh0 = _wkv7_bwd_kernel_with_mask(
            w, q, k, v, a, b, mask, dy, s, sa, dht
        )
        return dw, dq, dk, dv, da, db, dh0, None

    wk7_kernel_with_mask.defvjp(_fwd_with_mask, _bwd_with_mask)

    # ---------- 公共 API ----------
    def generalized_delta_rule(
        r: jnp.ndarray,
        w: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        initial_state: Optional[jnp.ndarray] = None,
        output_final_state: bool = True,
        head_first: bool = False,
        mask: Optional[jnp.ndarray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray]]:
        dtype = r.dtype
        r = _transpose_head(r, head_first)
        w = _transpose_head(w, head_first)
        k = _transpose_head(k, head_first)
        v = _transpose_head(v, head_first)
        a = _transpose_head(a, head_first)
        b = _transpose_head(b, head_first)

        B, T, H, K = r.shape
        if T % CHUNK_LEN:
            raise ValueError(
                f"Sequence length T={T} must be divisible by chunk_len={CHUNK_LEN}"
            )

        if initial_state is None:
            h0 = jnp.zeros((B, H, K, K), jnp.float32)
        else:
            h0 = jnp.asarray(initial_state, jnp.float32)

        if mask is None:
            out, last_state = wk7_kernel(w, r, k, v, a, b, h0)
        else:
            if mask.shape != (B, T):
                raise ValueError(
                    f"mask shape must be (B, T) = ({B}, {T}), got {mask.shape}"
                )
            mask = jnp.asarray(mask, jnp.bfloat16)
            out, last_state = wk7_kernel_with_mask(w, r, k, v, a, b, h0, mask)

        out = jnp.asarray(out, dtype)

        if output_final_state:
            return out, last_state
        return out

    # =========================================================================
    # 推理 Kernel (无 Mask)
    # =========================================================================
    def _wkv7_inference_kernel_impl(w, q, k, v, a, b, h0):
        B, T, H, K = q.shape
        dtype = q.dtype
        out_type = jax.ShapeDtypeStruct((B, T, H, K), dtype)
        s_type = jax.ShapeDtypeStruct((B, H, K, K), jnp.float32)

        y, s = jax.ffi.ffi_call(
            "wkv7_inference", (out_type, s_type), vmap_method="broadcast_all"
        )(w, q, k, v, a, b, h0)
        return y, s

    @custom_partitioning
    def _wkv7_inference_kernel(w, q, k, v, a, b, h0):
        return _wkv7_inference_kernel_impl(w, q, k, v, a, b, h0)

    _wkv7_inference_kernel.def_partition(
        infer_sharding_from_operands=_inf_infer_sharding,
        sharding_rule=INF_RULE,
        partition=_create_partition(_wkv7_inference_kernel_impl),
    )

    # =========================================================================
    # 推理 Kernel (带 Mask)
    # =========================================================================
    def _wkv7_inference_kernel_with_mask_impl(w, q, k, v, a, b, h0, mask):
        B, T, H, K = q.shape
        dtype = q.dtype
        out_type = jax.ShapeDtypeStruct((B, T, H, K), dtype)
        s_type = jax.ShapeDtypeStruct((B, H, K, K), jnp.float32)

        y, s = jax.ffi.ffi_call(
            "wkv7_inference_with_mask", (out_type, s_type), vmap_method="broadcast_all"
        )(w, q, k, v, a, b, mask, h0)
        return y, s

    @custom_partitioning
    def _wkv7_inference_kernel_with_mask(w, q, k, v, a, b, h0, mask):
        return _wkv7_inference_kernel_with_mask_impl(w, q, k, v, a, b, h0, mask)

    _wkv7_inference_kernel_with_mask.def_partition(
        infer_sharding_from_operands=_inf_infer_sharding,
        sharding_rule=INF_MASK_RULE,
        partition=_create_partition(_wkv7_inference_kernel_with_mask_impl),
    )

    # ---------- 公共推理 API ----------
    def generalized_delta_rule_inference(
        r: jnp.ndarray,
        w: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        output_final_state: bool = True,
        initial_state: Optional[jnp.ndarray] = None,
        head_first: bool = False,
        mask: Optional[jnp.ndarray] = None,
    ):
        dtype = r.dtype
        r = _transpose_head(r, head_first)
        w = _transpose_head(w, head_first)
        k = _transpose_head(k, head_first)
        v = _transpose_head(v, head_first)
        a = _transpose_head(a, head_first)
        b = _transpose_head(b, head_first)

        B, T, H, K = r.shape

        if initial_state is None:
            h0 = jnp.zeros((B, H, K, K), jnp.float32)
        else:
            h0 = jnp.asarray(initial_state, jnp.float32)

        if mask is None:
            out, final_state = _wkv7_inference_kernel(w, r, k, v, a, b, h0)
        else:
            if mask.shape != (B, T):
                raise ValueError(
                    f"mask shape must be (B, T) = ({B}, {T}), got {mask.shape}"
                )
            mask = jnp.asarray(mask, jnp.bfloat16)
            out, final_state = _wkv7_inference_kernel_with_mask(
                w, r, k, v, a, b, h0, mask
            )

        out = jnp.asarray(out, dtype)
        return (out, final_state) if output_final_state else out

    return [generalized_delta_rule, generalized_delta_rule_inference]



import os

os.environ["CUDA_VISIBLE_DEVICES"] = "1"
os.environ["KERAS_BACKEND"] = "tensorflow"
os.environ["KERNEL_TYPE"] = "cuda"

import numpy as np
from keras import ops


# ------------------------------------------------------------------
# 1. 构造输入
# ------------------------------------------------------------------
T = 1
B = 5
H = 6
K = 64
inputs = [np.random.randn(B, T, H, K) for _ in range(30)]
tf_inputs = [ops.convert_to_tensor(t, "bfloat16") for t in inputs]


def normalize(
    z,
    p=2,
    dim=-1,
    eps: float = 1e-12,
):
    # F.normalize like api
    denom = ops.norm(z, ord=p, axis=dim, keepdims=True)
    denom = ops.maximum(denom, 1e-12)
    return z / denom


mask = ops.concatenate(
    [ops.zeros((B, T // 2, 1, 1)), ops.ones((B, T // 2, 1, 1))], axis=1
)


a = -normalize(tf_inputs[3], dim=-1, p=2.0)
b = normalize(tf_inputs[3], dim=-1, p=2.0)

w = tf_inputs[4]  # decay / gate
r = tf_inputs[0]  # receptance
k = tf_inputs[1]
v = tf_inputs[2]
w = -ops.softplus(w) - 0.5
h0 = ops.convert_to_tensor(np.random.randn(B, H, K, K), "float32")
# ------------------------------------------------------------------
# 2. CUDA 版本前向 + 反向
# ------------------------------------------------------------------
from rwkv_ops import rwkv7_op_rnn


cuda_out, cuda_state = rwkv7_op_rnn(r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0)

r_n = ops.copy(r)
k_n = ops.copy(k)
v_n = ops.copy(v)
a_n = ops.copy(a)
b_n = ops.copy(b)
w_n = ops.copy(w)
h0_n = ops.copy(h0)

from rwkv_ops.rwkv7_kernel.native_keras_op import generalized_delta_rule

native_out, native_state = generalized_delta_rule(
    r=r_n, k=k_n, v=v_n, a=a_n, b=b_n, w=w_n, initial_state=h0_n
)


def test_is_close(name, x1, x2, atol=2.5e-2, rtol=1e-3):
    x1 = ops.convert_to_numpy(ops.cast(x1, "float32"))
    x2 = ops.convert_to_numpy(ops.cast(x2, "float32"))
    if np.sum(np.isnan(x1)) == 0 and np.sum(np.isnan(x2)) == 0:
        print(f"✅✅{name} 不存在nan✅✅")
    else:
        print(f"❌❌{name} 你妈的有nan❌❌")
    if np.sum(np.abs(x1 - x2)) < 1e-4:
        print(f"✅✅{name} 输出结果完全一致✅✅")
        return
    try:
        np.testing.assert_allclose(
            x1,
            x2,
            atol=atol,
            rtol=rtol,
        )
        print(f"✅ {name} 一致")
    except AssertionError as e:
        print(f"❌ {name} 不一致")
        print(e)


test_is_close("fwd_pred", native_out, cuda_out, atol=1e-5, rtol=1e-2)
test_is_close("fwd_state", native_state, cuda_state, atol=1e-5, rtol=1e-3)
print("🎉🎉🎉🎉test_script/test_tf_cuda_kernel_single.py测试结束🎉🎉🎉🎉")



import keras
from keras import ops


def get_mhc_kernel(KERNEL_TYPE="native"):
    from .native_op import (
        linear_and_reshape,
        mhc_post_op,
        mhc_pre_op_fused,
        mhc_rmsnorm,
    )

    if KERNEL_TYPE == "triton":
        if keras.config.backend() == "jax":
            import jax

            if jax.devices()[0].platform == "gpu":
                from .jax_triton_op.mhc_pre_op import mhc_pre_op_fused
                from .jax_triton_op.mhc_post_op import mhc_post_op
        elif keras.config.backend() == "torch":
            import torch

            if torch.cuda.is_available():
                from .torch_triton_op.mhc_pre_op import mhc_pre_op_fused
                from .torch_triton_op.mhc_post_op import mhc_post_op

    def mhc_pre_op(
        x,
        alpha_pre,
        alpha_post,
        alpha_res,
        phi,
        bias_pre,
        bias_post,
        bias_res,
        n,
        num_iters=20,
        eps=1e-8,
    ):
        """
        mHC 预处理算子（高层封装）

        功能：生成 mHC 核心层所需的完整输入，包括聚合后的单流特征、分发权重和双随机残差矩阵。
        实现：linear_and_reshape → Sinkhorn-Knopp → stream_aggregate

        --- 参数 ---
        x: [batch_size, seq_len, n , hidden_size], BF16/FP32
            多流输入特征

        alpha_pre, alpha_post, alpha_res: (1,), float32
            分支缩放系数

        phi: [n * hidden_size, M], BF16
            投影矩阵（M = n*(n+2)）

        bias: [M], float32
            线性偏置

        n: int
            流数量

        num_iters: int, default=20
            Sinkhorn-Knopp 迭代次数

        eps: float, default=1e-8
            Sinkhorn-Knopp 数值稳定常数

        --- 返回 ---
        x_layer_in: [batch_size, seq_len, hidden_size], 与 x.dtype 相同
            聚合后的单流特征，计算公式：sum(x * sigmoid(h_pre_raw), axis=-2)
            作为 Attention/FFN 等核心层的输入

        H_post: [batch_size, seq_len, n], 与 x.dtype 相同
            分发权重（已激活），用于将核心层输出广播回多流

        H_res: [batch_size, seq_len, n, n], 与 x.dtype 相同
            双随机残差矩阵，实现 n 个流之间的可逆混合
        """

        original_dtype = x.dtype
        B, T = ops.shape(x)[:2]
        x_flat = ops.reshape(x, [B, T, -1])
        x_norm = mhc_rmsnorm(x_flat, eps)
        h_pre_raw, h_post_raw, h_res_reshaped = linear_and_reshape(
            x_norm,
            alpha_pre,
            alpha_post,
            alpha_res,
            phi,
            bias_pre,
            bias_post,
            bias_res,
            n,
            eps=eps,
        )

        x_layer_in, H_res = mhc_pre_op_fused(
            x,
            h_res_reshaped,
            h_pre_raw,
            num_iters,
            eps,
        )

        return (
            ops.cast(x_layer_in, original_dtype),
            ops.cast(h_post_raw, "float32"),
            ops.cast(H_res, "float32"),
        )

    return mhc_pre_op, mhc_post_op



import os
import time
import jax
import jax.numpy as jnp
import numpy as np
from functools import partial

os.environ["KERNEL_TYPE"] = "triton"
os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["KERAS_BACKEND"] = "jax"
# 1. 导入你的 JAX-Triton 算子
# 假设已经按照之前的指南定义并导出
from rwkv_ops.mhc_kernel.jax_triton_op.mhc_pre_op import (
    mhc_pre_op_fused as triton_mhc_pre_op,
)
from rwkv_ops.mhc_kernel.native_op import mhc_pre_op_fused as native_mhc_pre_op

B, T, n, C = 128, 1024, 4, 512
dtype = jnp.bfloat16

# JAX 随机数管理
key = jax.random.PRNGKey(42)
k1, k2, k3 = jax.random.split(key, 3)


def get_inputs():
    x = jax.random.normal(k1, (B, T, n, C)).astype(jnp.bfloat16)
    h_res = jax.random.normal(k2, (B, T, n, n)).astype(jnp.float32)
    h_pre = jax.random.normal(k3, (B, T, n)).astype(jnp.float32)
    return x, h_res, h_pre


def test_is_close(name, x1, x2, atol=1e-2, rtol=1e-2):
    if x1 is None or x2 is None:
        print(f"❌ {name} 数据不存在")
        return

    x1_np = np.array(x1, dtype=np.float32)
    x2_np = np.array(x2, dtype=np.float32)

    error = np.abs(x1_np - x2_np)
    avg_error = error.mean()
    rel_error = avg_error / (np.abs(x1_np).mean() + 1e-9)

    print(f"[{name}]")
    print(f"  平均绝对误差: {avg_error:.6e}")
    print(f"  平均相对误差: {rel_error:.6e}")

    # 计算完全一样的比例 (在 BF16 下通常较低，主要看 max_diff)
    equal_rate = np.sum(error < 1e-6) / x1_np.size * 100
    print(f"  {equal_rate:.2f}% 的数据在 1e-6 误差内一致")

    if np.isnan(x1_np).any() or np.isnan(x2_np).any():
        print("  ❌❌ 存在 NaN ❌❌")
        return

    try:
        np.testing.assert_allclose(x1_np, x2_np, atol=atol, rtol=rtol)
        print(f"  ✅ 数值一致 (max_diff: {error.max():.6e})")
    except AssertionError:
        print("  ❌ 数值不一致")


print("--- 开始数值正确性校验 ---")

x_data, hr_data, hp_data = get_inputs()


# 定义 Loss 函数用于测试反向梯度
def loss_fn(op_func, x, hr, hp):
    out_x, out_h = op_func(x, hr, hp, num_iters=20)
    return jnp.mean(out_x.astype(jnp.float32) ** 2) + jnp.mean(out_h**2)


# --- 前向对比 ---
out_x_n, out_h_n = native_mhc_pre_op(x_data, hr_data, hp_data)
out_x_t, out_h_t = triton_mhc_pre_op(x_data, hr_data, hp_data)

test_is_close("Forward: x_layer_in", out_x_n, out_x_t)
test_is_close("Forward: H_res", out_h_n, out_h_t)

# --- 反向梯度对比 ---
grad_native_fn = jax.grad(partial(loss_fn, native_mhc_pre_op), argnums=(0, 1, 2))
grad_triton_fn = jax.grad(partial(loss_fn, triton_mhc_pre_op), argnums=(0, 1, 2))

print("\n--- 开始梯度对比 ---")
gx_n, ghr_n, ghp_n = grad_native_fn(x_data, hr_data, hp_data)
gx_t, ghr_t, ghp_t = grad_triton_fn(x_data, hr_data, hp_data)

test_is_close("Gradient: x", gx_n, gx_t)
test_is_close("Gradient: h_res_in", ghr_n, ghr_t)
test_is_close("Gradient: h_pre_in", ghp_n, ghp_t)

print("\n" + "=" * 40)
print("🚀 开始全流程性能基准测试 (Forward + Backward)")
print("=" * 40)

n_warmup = 10
n_repeat = 100

# JAX 需要显式 jit
jit_native_full = jax.jit(
    jax.value_and_grad(partial(loss_fn, native_mhc_pre_op), argnums=(0, 1, 2))
)
jit_triton_full = jax.jit(
    jax.value_and_grad(partial(loss_fn, triton_mhc_pre_op), argnums=(0, 1, 2))
)


def benchmark_all(run_fn, name, x, hr, hp):
    # 预热
    for _ in range(n_warmup):
        res = run_fn(x, hr, hp)
        jax.block_until_ready(res)

    start_time = time.perf_counter()
    for _ in range(n_repeat):
        res = run_fn(x, hr, hp)
        jax.block_until_ready(res)
    end_time = time.perf_counter()

    avg_time_ms = (end_time - start_time) / n_repeat * 1000
    print(f"👉 {name:20s} Full (Fwd+Bwd): {avg_time_ms:.4f} ms")
    return avg_time_ms


# 运行 Benchmark
print(f"配置: B={B}, T={T}, n={n}, C={C}, dtype={dtype}")

# 1. Native Eager (JAX 极慢，不建议跑太大型的 B, T)
# t_native_eager = benchmark_all(lambda x, hr, hp: jax.grad(partial(loss_fn, native_mhc_pre_op))(x, hr, hp), "Native (Eager)", x_data, hr_data, hp_data)

t_native_jit = benchmark_all(
    jax.jit(native_mhc_pre_op), "JAX JIT (Native)", x_data, hr_data, hp_data
)

# 3. Triton Fused JIT
t_triton_jit = benchmark_all(
    jax.jit(triton_mhc_pre_op), "JAX-Triton Fused", x_data, hr_data, hp_data
)

print("\n📊 性能提升总结:")
print(f"Triton vs JIT Native前向: {t_native_jit / t_triton_jit:.2f}x faster")
t_native_jit = benchmark_all(
    jit_native_full, "JAX JIT (Native)", x_data, hr_data, hp_data
)

# 3. Triton Fused JIT
t_triton_jit = benchmark_all(
    jit_triton_full, "JAX-Triton Fused", x_data, hr_data, hp_data
)


print(f"Triton vs JIT Native: {t_native_jit / t_triton_jit:.2f}x faster")
# if 't_native_eager' in locals():
#     print(f"Triton vs Eager:      {t_native_eager / t_triton_jit:.2f}x faster")
print("\n🎉🎉 MHC Pre-Op 全校验结束 🎉🎉")



import triton
import triton.language as tl


# ======================================================================================
#                                   无 Mask 前向传播
# ======================================================================================
@triton.autotune(
    configs=[
        # 【稳定性修复 1/2】：强制 MINI_BSZ=1
        # 核心原因：限制每个 GPU Block 只处理一个 Batch 样本，避免因 MINI_BSZ > 1
        # 导致在 AMD 显卡上出现超大的寄存器/共享内存分配，从而绕开 ROCm 编译器的
        # 寄存器溢出 (Register Spilling) Bug，这是导致 Segfault 的主要原因之一。
        # 这种做法牺牲了极小的块内并行度，但换来了跨平台的绝对稳定性。
        triton.Config({"MINI_BSZ": 1}, num_warps=num_warps, num_stages=num_stages)
        # 补偿策略：增加 num_warps 和 num_stages 可以利用更多线程和更深的流水线来
        # 隐藏内存延迟，弥补 MINI_BSZ=1 带来的性能损失。
        for num_warps in [4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["H_SIZE"],
)
@triton.jit
def rwkv7_fwd_kernel(
    R,
    W,
    K,
    V,
    A,
    B_param,
    H0,
    B_BATCH,
    N_HEAD,
    T_LEN,
    OUT,
    SA_OUT,
    STATE_CHKP,
    H_SIZE: tl.constexpr,
    CHUNK_LEN: tl.constexpr,
    MINI_BSZ: tl.constexpr,
):
    pid_b = tl.program_id(0)
    pid_h = tl.program_id(1)

    # --- 1. 计算 Batch 掩码 ---
    b_range = pid_b * MINI_BSZ + tl.arange(0, MINI_BSZ)
    b_mask = b_range < B_BATCH
    cols = tl.arange(0, H_SIZE)

    # 【稳定性修复 2/2】：强制使用 64 位整数进行指针运算
    # 核心原因：防止在 Batch, T_LEN, N_HEAD, H_SIZE 较大时发生 32 位整数溢出，
    # 导致计算出错误的负数内存地址。同时，这也能规避 AMD 编译器在处理
    # 32 位到 64 位地址扩展时可能存在的 Bug。
    b_range_i64 = b_range.to(tl.int64)
    N_HEAD_i64 = tl.cast(N_HEAD, tl.int64)
    T_LEN_i64 = tl.cast(T_LEN, tl.int64)
    H_SIZE_i64 = tl.cast(H_SIZE, tl.int64)
    pid_h_i64 = tl.cast(pid_h, tl.int64)
    CHUNK_LEN_i64 = tl.cast(CHUNK_LEN, tl.int64)

    # --- 2. 构造各种形状的掩码 ---
    ptr_mask = b_mask[:, None]
    state_mask = b_mask[:, None, None]

    # --- 3. 计算基础偏移 (使用 64 位整数) ---
    base_ptr_off = (
        (b_range_i64[:, None] * N_HEAD_i64 * T_LEN_i64 * H_SIZE_i64)
        + (pid_h_i64 * T_LEN_i64 * H_SIZE_i64)
        + cols[None, :]
    )
    h0_base = (b_range_i64[:, None, None] * N_HEAD_i64 * H_SIZE_i64 * H_SIZE_i64) + (
        pid_h_i64 * H_SIZE_i64 * H_SIZE_i64
    )

    # --- 4. 初始化状态加载 ---
    row_idx = tl.arange(0, H_SIZE)[None, :, None]
    col_idx = tl.arange(0, H_SIZE)[None, None, :]
    state = tl.load(
        H0 + h0_base + row_idx * H_SIZE + col_idx, mask=state_mask, other=0.0
    ).to(tl.float32)

    for t in range(0, T_LEN):
        t_off = t * H_SIZE

        # a. 向量加载
        rv = tl.load(R + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        wv = tl.load(W + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        kv = tl.load(K + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        vv = tl.load(V + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        av = tl.load(A + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        bv = tl.load(B_param + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )

        # b. 数学计算
        w_decay = tl.exp(-tl.exp(wv))
        sa_vec = tl.sum(state * av[:, None, :], axis=2)
        tl.store(
            SA_OUT + base_ptr_off + t_off,
            sa_vec.to(SA_OUT.dtype.element_ty),
            mask=ptr_mask,
        )

        # c. 状态演进
        state = (
            state * w_decay[:, None, :]
            + sa_vec[:, :, None] * bv[:, None, :]
            + vv[:, :, None] * kv[:, None, :]
        )

        # d. 计算输出
        y_vec = tl.sum(state * rv[:, None, :], axis=2)
        tl.store(
            OUT + base_ptr_off + t_off, y_vec.to(OUT.dtype.element_ty), mask=ptr_mask
        )

        # e. 状态快照
        if (t + 1) % CHUNK_LEN == 0:
            chkp_t = (tl.cast(t, tl.int64) + 1) // CHUNK_LEN_i64 - 1
            chunk_num_i64 = T_LEN_i64 // CHUNK_LEN_i64
            chkp_base = (
                (
                    b_range_i64[:, None, None]
                    * N_HEAD_i64
                    * chunk_num_i64
                    * H_SIZE_i64
                    * H_SIZE_i64
                )
                + (pid_h_i64 * chunk_num_i64 * H_SIZE_i64 * H_SIZE_i64)
                + (chkp_t * H_SIZE_i64 * H_SIZE_i64)
            )
            tl.store(
                STATE_CHKP + chkp_base + row_idx * H_SIZE + col_idx,
                state,
                mask=state_mask,
            )


# ======================================================================================
#                                   无 Mask 反向传播
# ======================================================================================
@triton.autotune(
    configs=[
        triton.Config({"MINI_BSZ": 1}, num_warps=num_warps, num_stages=num_stages)
        for num_warps in [4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["H_SIZE"],
)
@triton.jit
def rwkv7_bwd_kernel(
    R,
    W,
    K,
    V,
    A,
    B_param,
    SA,
    STATE_CHKP,
    B_BATCH,
    N_HEAD,
    T_LEN,
    DY,
    DHT,
    DR,
    DW,
    DK,
    DV,
    DA,
    DB,
    DH0,
    H_SIZE: tl.constexpr,
    CHUNK_LEN: tl.constexpr,
    MINI_BSZ: tl.constexpr,
):
    pid_b = tl.program_id(0)
    pid_h = tl.program_id(1)

    b_range = pid_b * MINI_BSZ + tl.arange(0, MINI_BSZ)
    b_mask = b_range < B_BATCH
    cols = tl.arange(0, H_SIZE)
    row_idx = tl.arange(0, H_SIZE)[None, :, None]
    col_idx = tl.arange(0, H_SIZE)[None, None, :]

    b_range_i64 = b_range.to(tl.int64)
    N_HEAD_i64 = tl.cast(N_HEAD, tl.int64)
    T_LEN_i64 = tl.cast(T_LEN, tl.int64)
    H_SIZE_i64 = tl.cast(H_SIZE, tl.int64)
    pid_h_i64 = tl.cast(pid_h, tl.int64)
    CHUNK_LEN_i64 = tl.cast(CHUNK_LEN, tl.int64)

    ptr_mask = b_mask[:, None]
    state_mask = b_mask[:, None, None]

    base_ptr_off = (
        (b_range_i64[:, None] * N_HEAD_i64 * T_LEN_i64 * H_SIZE_i64)
        + (pid_h_i64 * T_LEN_i64 * H_SIZE_i64)
        + cols[None, :]
    )
    dht_base = (b_range_i64[:, None, None] * N_HEAD_i64 * H_SIZE_i64 * H_SIZE_i64) + (
        pid_h_i64 * H_SIZE_i64 * H_SIZE_i64
    )

    dS = tl.load(
        DHT + dht_base + row_idx * H_SIZE + col_idx, mask=state_mask, other=0.0
    ).to(tl.float32)
    S_t = tl.zeros([MINI_BSZ, H_SIZE, H_SIZE], dtype=tl.float32)

    for t in range(T_LEN - 1, -1, -1):
        t_off = t * H_SIZE

        rv = tl.load(R + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        wv = tl.load(W + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        kv = tl.load(K + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        vv = tl.load(V + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        av = tl.load(A + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        bv = tl.load(B_param + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )
        dyv = tl.load(DY + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )
        sav = tl.load(SA + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )

        w_decay = tl.exp(-tl.exp(wv))
        w_grad_factor = w_decay * (-tl.exp(wv))

        if (t + 1) % CHUNK_LEN == 0:
            chkp_t = (tl.cast(t, tl.int64) + 1) // CHUNK_LEN_i64 - 1
            chunk_num_i64 = T_LEN_i64 // CHUNK_LEN_i64
            chkp_base = (
                (
                    b_range_i64[:, None, None]
                    * N_HEAD_i64
                    * chunk_num_i64
                    * H_SIZE_i64
                    * H_SIZE_i64
                )
                + (pid_h_i64 * chunk_num_i64 * H_SIZE_i64 * H_SIZE_i64)
                + (chkp_t * H_SIZE_i64 * H_SIZE_i64)
            )
            S_t = tl.load(
                STATE_CHKP + chkp_base + row_idx * H_SIZE + col_idx,
                mask=state_mask,
                other=0.0,
            ).to(tl.float32)

        dr = tl.sum(S_t * dyv[:, :, None], axis=1)
        tl.store(DR + base_ptr_off + t_off, dr.to(DR.dtype.element_ty), mask=ptr_mask)

        inv_w = 1.0 / (w_decay + 1e-6)
        S_t = (
            S_t - vv[:, :, None] * kv[:, None, :] - sav[:, :, None] * bv[:, None, :]
        ) * inv_w[:, None, :]

        dS = dS + dyv[:, :, None] * rv[:, None, :]

        dw = tl.sum(dS * S_t, axis=1) * w_grad_factor
        dk = tl.sum(dS * vv[:, :, None], axis=1)
        dv = tl.sum(dS * kv[:, None, :], axis=2)
        db = tl.sum(dS * sav[:, :, None], axis=1)
        dsa = tl.sum(dS * bv[:, None, :], axis=2)
        da = tl.sum(S_t * dsa[:, :, None], axis=1)

        tl.store(DW + base_ptr_off + t_off, dw.to(DW.dtype.element_ty), mask=ptr_mask)
        tl.store(DK + base_ptr_off + t_off, dk.to(DK.dtype.element_ty), mask=ptr_mask)
        tl.store(DV + base_ptr_off + t_off, dv.to(DV.dtype.element_ty), mask=ptr_mask)
        tl.store(DB + base_ptr_off + t_off, db.to(DB.dtype.element_ty), mask=ptr_mask)
        tl.store(DA + base_ptr_off + t_off, da.to(DA.dtype.element_ty), mask=ptr_mask)

        dS = dS * w_decay[:, None, :] + dsa[:, :, None] * av[:, None, :]

        if t == 0:
            tl.store(
                DH0 + dht_base + row_idx * H_SIZE + col_idx,
                dS.to(DH0.dtype.element_ty),
                mask=state_mask,
            )


# ======================================================================================
#                                   带 Mask 前向传播
# ======================================================================================
@triton.autotune(
    configs=[
        triton.Config({"MINI_BSZ": 1}, num_warps=num_warps, num_stages=num_stages)
        for num_warps in [4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["H_SIZE"],
)
@triton.jit
def rwkv7_fwd_kernel_with_mask(
    R,
    W,
    K,
    V,
    A,
    B_param,
    MASK,
    H0,
    B_BATCH,
    N_HEAD,
    T_LEN,
    OUT,
    SA_OUT,
    STATE_CHKP,
    H_SIZE: tl.constexpr,
    CHUNK_LEN: tl.constexpr,
    MINI_BSZ: tl.constexpr,
):
    pid_b = tl.program_id(0)
    pid_h = tl.program_id(1)

    b_range = pid_b * MINI_BSZ + tl.arange(0, MINI_BSZ)
    b_mask = b_range < B_BATCH
    cols = tl.arange(0, H_SIZE)

    b_range_i64 = b_range.to(tl.int64)
    N_HEAD_i64 = tl.cast(N_HEAD, tl.int64)
    T_LEN_i64 = tl.cast(T_LEN, tl.int64)
    H_SIZE_i64 = tl.cast(H_SIZE, tl.int64)
    pid_h_i64 = tl.cast(pid_h, tl.int64)
    CHUNK_LEN_i64 = tl.cast(CHUNK_LEN, tl.int64)

    ptr_mask = b_mask[:, None]
    state_mask = b_mask[:, None, None]

    base_ptr_off = (
        (b_range_i64[:, None] * N_HEAD_i64 * T_LEN_i64 * H_SIZE_i64)
        + (pid_h_i64 * T_LEN_i64 * H_SIZE_i64)
        + cols[None, :]
    )
    mask_ptr_base = b_range_i64 * T_LEN_i64

    h0_base = (b_range_i64[:, None, None] * N_HEAD_i64 * H_SIZE_i64 * H_SIZE_i64) + (
        pid_h_i64 * H_SIZE_i64 * H_SIZE_i64
    )
    row_idx = tl.arange(0, H_SIZE)[None, :, None]
    col_idx = tl.arange(0, H_SIZE)[None, None, :]

    state = tl.load(
        H0 + h0_base + row_idx * H_SIZE + col_idx, mask=state_mask, other=0.0
    ).to(tl.float32)

    for t in range(0, T_LEN):
        t_off = t * H_SIZE

        rv = tl.load(R + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        wv = tl.load(W + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        kv = tl.load(K + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        vv = tl.load(V + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        av = tl.load(A + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        bv = tl.load(B_param + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )

        m_val = tl.load(MASK + mask_ptr_base + t, mask=b_mask, other=0.0).to(tl.float32)
        m_3d = m_val[:, None, None]

        w_decay = tl.exp(-tl.exp(wv))
        sa_vec = tl.sum(state * av[:, None, :], axis=2)
        tl.store(
            SA_OUT + base_ptr_off + t_off,
            sa_vec.to(SA_OUT.dtype.element_ty),
            mask=ptr_mask,
        )

        state_cand = (
            state * w_decay[:, None, :]
            + sa_vec[:, :, None] * bv[:, None, :]
            + vv[:, :, None] * kv[:, None, :]
        )

        y_vec = tl.sum(state_cand * rv[:, None, :], axis=2)
        tl.store(
            OUT + base_ptr_off + t_off, y_vec.to(OUT.dtype.element_ty), mask=ptr_mask
        )

        state = m_3d * state_cand + (1.0 - m_3d) * state

        if (t + 1) % CHUNK_LEN == 0:
            chkp_t = (tl.cast(t, tl.int64) + 1) // CHUNK_LEN_i64 - 1
            chunk_num_i64 = T_LEN_i64 // CHUNK_LEN_i64
            chkp_base = (
                (
                    b_range_i64[:, None, None]
                    * N_HEAD_i64
                    * chunk_num_i64
                    * H_SIZE_i64
                    * H_SIZE_i64
                )
                + (pid_h_i64 * chunk_num_i64 * H_SIZE_i64 * H_SIZE_i64)
                + (chkp_t * H_SIZE_i64 * H_SIZE_i64)
            )
            tl.store(
                STATE_CHKP + chkp_base + row_idx * H_SIZE + col_idx,
                state,
                mask=state_mask,
            )


# ======================================================================================
#                                   带 Mask 反向传播
# ======================================================================================
@triton.autotune(
    configs=[
        triton.Config({"MINI_BSZ": 1}, num_warps=num_warps, num_stages=num_stages)
        for num_warps in [4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["H_SIZE"],
)
@triton.jit
def rwkv7_bwd_kernel_with_mask(
    R,
    W,
    K,
    V,
    A,
    B_param,
    MASK,
    SA,
    STATE_CHKP,
    B_BATCH,
    N_HEAD,
    T_LEN,
    DY,
    DHT,
    DR,
    DW,
    DK,
    DV,
    DA,
    DB,
    DH0,
    H_SIZE: tl.constexpr,
    CHUNK_LEN: tl.constexpr,
    MINI_BSZ: tl.constexpr,
):
    pid_b = tl.program_id(0)
    pid_h = tl.program_id(1)

    b_range = pid_b * MINI_BSZ + tl.arange(0, MINI_BSZ)
    b_mask = b_range < B_BATCH

    cols = tl.arange(0, H_SIZE)
    row_idx = tl.arange(0, H_SIZE)[None, :, None]
    col_idx = tl.arange(0, H_SIZE)[None, None, :]

    b_range_i64 = b_range.to(tl.int64)
    N_HEAD_i64 = tl.cast(N_HEAD, tl.int64)
    T_LEN_i64 = tl.cast(T_LEN, tl.int64)
    H_SIZE_i64 = tl.cast(H_SIZE, tl.int64)
    pid_h_i64 = tl.cast(pid_h, tl.int64)
    CHUNK_LEN_i64 = tl.cast(CHUNK_LEN, tl.int64)

    ptr_mask = b_mask[:, None]
    state_mask = b_mask[:, None, None]

    base_ptr_off = (
        (b_range_i64[:, None] * N_HEAD_i64 * T_LEN_i64 * H_SIZE_i64)
        + (pid_h_i64 * T_LEN_i64 * H_SIZE_i64)
        + cols[None, :]
    )
    mask_ptr_base = b_range_i64 * T_LEN_i64

    dht_base = (b_range_i64[:, None, None] * N_HEAD_i64 * H_SIZE_i64 * H_SIZE_i64) + (
        pid_h_i64 * H_SIZE_i64 * H_SIZE_i64
    )
    dS = tl.load(
        DHT + dht_base + row_idx * H_SIZE + col_idx, mask=state_mask, other=0.0
    ).to(tl.float32)

    S_t = tl.zeros([MINI_BSZ, H_SIZE, H_SIZE], dtype=tl.float32)

    for t in range(T_LEN - 1, -1, -1):
        t_off = t * H_SIZE

        rv = tl.load(R + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        wv = tl.load(W + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        kv = tl.load(K + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        vv = tl.load(V + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        av = tl.load(A + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(tl.float32)
        bv = tl.load(B_param + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )
        dyv = tl.load(DY + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )
        sav = tl.load(SA + base_ptr_off + t_off, mask=ptr_mask, other=0.0).to(
            tl.float32
        )

        m_val = tl.load(MASK + mask_ptr_base + t, mask=b_mask, other=0.0).to(tl.float32)
        m_3d = m_val[:, None, None]

        w_decay = tl.exp(-tl.exp(wv))
        w_grad_factor = w_decay * (-tl.exp(wv))

        if (t + 1) % CHUNK_LEN == 0:
            chkp_t = (tl.cast(t, tl.int64) + 1) // CHUNK_LEN_i64 - 1
            chunk_num_i64 = T_LEN_i64 // CHUNK_LEN_i64
            chkp_base = (
                (
                    b_range_i64[:, None, None]
                    * N_HEAD_i64
                    * chunk_num_i64
                    * H_SIZE_i64
                    * H_SIZE_i64
                )
                + (pid_h_i64 * chunk_num_i64 * H_SIZE_i64 * H_SIZE_i64)
                + (chkp_t * H_SIZE_i64 * H_SIZE_i64)
            )
            S_t = tl.load(
                STATE_CHKP + chkp_base + row_idx * H_SIZE + col_idx,
                mask=state_mask,
                other=0.0,
            ).to(tl.float32)

        s_cand_for_dr = (
            S_t * w_decay[:, None, :]
            + sav[:, :, None] * bv[:, None, :]
            + vv[:, :, None] * kv[:, None, :]
        )
        s_for_dr = m_3d * S_t + (1.0 - m_3d) * s_cand_for_dr
        dr = tl.sum(s_for_dr * dyv[:, :, None], axis=1)
        tl.store(DR + base_ptr_off + t_off, dr.to(DR.dtype.element_ty), mask=ptr_mask)

        dS_curr = dyv[:, :, None] * rv[:, None, :]
        dS = dS + dS_curr
        dS_old = dS

        inv_w = 1.0 / (w_decay + 1e-6)
        S_prev = (
            S_t - vv[:, :, None] * kv[:, None, :] - sav[:, :, None] * bv[:, None, :]
        ) * inv_w[:, None, :]
        S_t = m_3d * S_prev + (1.0 - m_3d) * S_t

        dS_param = m_3d * dS_old + (1.0 - m_3d) * dS_curr

        dw = tl.sum(dS_param * S_t, axis=1) * w_grad_factor
        dk = tl.sum(dS_param * vv[:, :, None], axis=1)
        dv = tl.sum(dS_param * kv[:, None, :], axis=2)
        db = tl.sum(dS_param * sav[:, :, None], axis=1)
        dsa = tl.sum(dS_param * bv[:, None, :], axis=2)
        da = tl.sum(S_t * dsa[:, :, None], axis=1)

        tl.store(DW + base_ptr_off + t_off, dw.to(DW.dtype.element_ty), mask=ptr_mask)
        tl.store(DK + base_ptr_off + t_off, dk.to(DK.dtype.element_ty), mask=ptr_mask)
        tl.store(DV + base_ptr_off + t_off, dv.to(DV.dtype.element_ty), mask=ptr_mask)
        tl.store(DB + base_ptr_off + t_off, db.to(DB.dtype.element_ty), mask=ptr_mask)
        tl.store(DA + base_ptr_off + t_off, da.to(DA.dtype.element_ty), mask=ptr_mask)

        trans = dS_param * w_decay[:, None, :] + dsa[:, :, None] * av[:, None, :]
        penetration = dS_old - dS_param
        dS = trans + (1.0 - m_3d) * penetration

        if t == 0:
            tl.store(
                DH0 + dht_base + row_idx * H_SIZE + col_idx,
                dS.to(DH0.dtype.element_ty),
                mask=state_mask,
            )



import os

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["KERAS_BACKEND"] = "torch"
os.environ["KERNEL_TYPE"] = "triton"

import numpy as np
import torch
from torch.nn import functional as F
from keras import ops

# ------------------------------------------------------------------
# 1. 构造输入
# ------------------------------------------------------------------
T = 1024
B = 5
H = 6
K = 64
inputs = [np.random.randn(B, T, H, K) for _ in range(30)]
torch_inputs = [torch.from_numpy(t).bfloat16().cuda() for t in inputs]

a = -F.normalize(torch_inputs[3], dim=-1, p=2.0)
b = F.normalize(torch_inputs[3], dim=-1, p=2.0)

w = torch_inputs[4]  # decay / gate
r = torch_inputs[0]  # receptance
k = torch_inputs[1]
v = torch_inputs[2]
w = -ops.softplus(w) - 0.5
h0 = torch.from_numpy(np.random.randn(B, H, K, K)).float().cuda()
# ------------------------------------------------------------------
# 2. CUDA 版本前向 + 反向
# ------------------------------------------------------------------
from rwkv_ops import rwkv7_op  # 即 get_generalized_delta_rule 返回的函数

# 要求计算梯度，必须设置 requires_grad
for t in [r, k, v, a, b, w, h0]:
    t.requires_grad_(True)

cuda_out, cuda_state = rwkv7_op(
    r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0, output_final_state=True
)


# ------------------------------------------------------------------
# 3. Native 版本前向 + 反向
# ------------------------------------------------------------------
from rwkv_ops.rwkv7_kernel.native_keras_op import generalized_delta_rule

# 重新构造一份 detach 的新张量，避免梯度混淆
r_n = r.detach().clone().requires_grad_(True)
k_n = k.detach().clone().requires_grad_(True)
v_n = v.detach().clone().requires_grad_(True)
a_n = a.detach().clone().requires_grad_(True)
b_n = b.detach().clone().requires_grad_(True)
w_n = w.detach().clone().requires_grad_(True)
h0_n = h0.detach().clone().requires_grad_(True)

native_out, native_state = generalized_delta_rule(
    r=r_n, k=k_n, v=v_n, a=a_n, b=b_n, w=w_n, initial_state=h0_n
)


# ------------------------------------------------------------------
# 4. 前向结果比较
# ------------------------------------------------------------------
def test_is_close(name, x1, x2, atol=5e-3, rtol=1e-3):
    t1 = ops.convert_to_numpy(ops.cast(x1, "float32"))
    t2 = ops.convert_to_numpy(ops.cast(x2, "float32"))
    error = np.abs(t1 - t2)
    max_diff = error.max()
    avg_error = error.mean()
    rel_error = avg_error / (np.abs(t1).mean() + 1e-9)
    print("-" * 100)
    print(f"[{name}]")
    print(f"  平均绝对误差: {avg_error:.6e}")
    print(f"  平均相对误差: {rel_error:.6e}")

    equal_rate = np.sum(error < 1e-7) / t1.size * 100
    print(f"  {equal_rate:.2f}% 的数据完全一样")
    try:
        np.testing.assert_allclose(t1, t2, atol=atol, rtol=rtol)
        print(f"✅ {name} 一致 (Max Diff: {max_diff:.6e})")
        return True
    except AssertionError as e:
        print(f"❌ {name} 不一致! Max Diff: {max_diff:.6e}")
        print(e)
        return False


test_is_close("fwd_pred", native_out, cuda_out, atol=1e-5, rtol=1e-2)
test_is_close("fwd_state", native_state, cuda_state, atol=1e-5, rtol=1e-3)
print("前向测试完毕")

loss_native = ((native_out).pow(2).mean() - (native_state).pow(2).mean()).abs()
loss_native.backward()

loss_cuda = ((cuda_out).pow(2).mean() - (cuda_state).pow(2).mean()).abs()
loss_cuda.backward()
# ------------------------------------------------------------------
# 5. 梯度比较
# -----------------------------------w -------------------------------
grad_names = ["r", "k", "v", "a", "b", "w", "h0"]
cuda_grads = [t.grad.float() for t in [r, k, v, a, b, w, h0]]
native_grads = [t.grad.float() for t in [r_n, k_n, v_n, a_n, b_n, w_n, h0_n]]

for name, g_cuda, g_native in zip(grad_names, cuda_grads, native_grads):
    test_is_close(name, g_cuda, g_native, atol=7e-3)
print("🎉🎉🎉🎉test_script/test_torch_cuda_kernel.py测试结束🎉🎉🎉🎉")

# ------------------------------------------------------------------
# 6. 带 Mask 版本测试
# ------------------------------------------------------------------
print("\n" + "=" * 50 + " 带 Mask 版本测试 " + "=" * 50)

# 构造 Mask：随机混合 0 和 1，模拟部分更新场景
# 形状 [B, T]，值 0.0 或 1.0
mask = torch.ones(B, T, device="cuda", dtype=torch.float32)
# 随机冻结某些位置（例如 30% 的时间步不更新状态）
freeze_indices = torch.rand(B, T) < 0.3
mask[freeze_indices] = 0.0

print(f"Mask 冻结比例: {(mask == 0).float().mean().item():.2%}")

# 重新构造输入（避免与之前测试的梯度混淆）
r_m = r.detach().clone().requires_grad_(True)
k_m = k.detach().clone().requires_grad_(True)
v_m = v.detach().clone().requires_grad_(True)
a_m = a.detach().clone().requires_grad_(True)
b_m = b.detach().clone().requires_grad_(True)
w_m = w.detach().clone().requires_grad_(True)
h0_m = h0.detach().clone().requires_grad_(True)

r_m_n = r.detach().clone().requires_grad_(True)
k_m_n = k.detach().clone().requires_grad_(True)
v_m_n = v.detach().clone().requires_grad_(True)
a_m_n = a.detach().clone().requires_grad_(True)
b_m_n = b.detach().clone().requires_grad_(True)
w_m_n = w.detach().clone().requires_grad_(True)
h0_m_n = h0.detach().clone().requires_grad_(True)
mask[:, -5:] = 0
# ------------------------------------------------------------------
# 7. CUDA 带 Mask 前向 + 反向
# ------------------------------------------------------------------
cuda_out_mask, cuda_state_mask = rwkv7_op(
    r=r_m,
    k=k_m,
    v=v_m,
    a=a_m,
    b=b_m,
    w=w_m,
    mask=mask,  # 传入 mask
    initial_state=h0_m,
    output_final_state=True,
)

# ------------------------------------------------------------------
# 8. Native 带 Mask 前向 + 反向
# ------------------------------------------------------------------


native_out_mask, native_state_mask = generalized_delta_rule(
    r=r_m_n,
    k=k_m_n,
    v=v_m_n,
    a=a_m_n,
    b=b_m_n,
    w=w_m_n,
    mask=mask,  # 传入 mask
    initial_state=h0_m_n,
)

# ------------------------------------------------------------------
# 9. 带 Mask 前向结果比较
# ------------------------------------------------------------------
print("\n>>> 带 Mask 前向对比")
test_is_close("fwd_pred_mask", native_out_mask, cuda_out_mask, atol=1e-5, rtol=1e-2)
test_is_close(
    "fwd_state_mask", native_state_mask, cuda_state_mask, atol=1e-5, rtol=1e-3
)

# ------------------------------------------------------------------
# 10. 带 Mask 梯度比较
# ------------------------------------------------------------------
loss_cuda_mask = (cuda_out_mask.pow(2).mean() - cuda_state_mask.pow(2).mean()).abs()
loss_cuda_mask.backward()

loss_native_mask = (
    native_out_mask.pow(2).mean() - native_state_mask.pow(2).mean()
).abs()
loss_native_mask.backward()

print("\n>>> 带 Mask 梯度对比")
grad_names = ["r", "k", "v", "a", "b", "w", "h0"]
cuda_grads_mask = [t.grad.float() for t in [r_m, k_m, v_m, a_m, b_m, w_m, h0_m]]
native_grads_mask = [
    t.grad.float() for t in [r_m_n, k_m_n, v_m_n, a_m_n, b_m_n, w_m_n, h0_m_n]
]

for name, g_cuda, g_native in zip(grad_names, cuda_grads_mask, native_grads_mask):
    test_is_close(f"{name}_mask", g_cuda, g_native, atol=7e-3)

# ------------------------------------------------------------------
# 11. Mask 物理一致性验证（关键测试）
# ------------------------------------------------------------------
print("\n>>> Mask 物理一致性验证")

# 测试：全 0 Mask 时状态应该完全不变
mask_all_zero = torch.zeros(B, T, device="cuda", dtype=torch.float32)
h0_frozen = h0.detach().clone()

with torch.no_grad():
    out_frozen, state_frozen = rwkv7_op(
        r=r,
        k=k,
        v=v,
        a=a,
        b=b,
        w=w,
        mask=mask_all_zero,
        initial_state=h0_frozen,
        output_final_state=True,
    )

    diff = (state_frozen - h0_frozen).abs().max().item()
    if diff < 1e-6:
        print(f"✅ 全 0 Mask 状态冻结验证通过 (Max Diff: {diff:.2e})")
    else:
        print(f"❌ 全 0 Mask 状态错误改变 (Max Diff: {diff:.2e})")

# 测试：全 1 Mask 时应该与无 Mask 版本等价
mask_all_one = torch.ones(B, T, device="cuda", dtype=torch.float32)

with torch.no_grad():
    out_one_mask, state_one_mask = rwkv7_op(
        r=r,
        k=k,
        v=v,
        a=a,
        b=b,
        w=w,
        mask=mask_all_one,
        initial_state=h0,
        output_final_state=True,
    )

    # 与之前无 mask 的 cuda_out/cuda_state 比较
    pred_diff = (out_one_mask - cuda_out).abs().max().item()
    state_diff = (state_one_mask - cuda_state).abs().max().item()

    if pred_diff < 1e-5 and state_diff < 1e-5:
        print(
            f"✅ 全 1 Mask 与无 Mask 等价 (Pred Diff: {pred_diff:.2e}, State Diff: {state_diff:.2e})"
        )
    else:
        print(
            f"⚠️ 全 1 Mask 与无 Mask 存在差异 (Pred Diff: {pred_diff:.2e}, State Diff: {state_diff:.2e})"
        )

print("\n" + "=" * 50 + " Mask 测试结束 " + "=" * 50)



import torch
import triton

# 假设你的 kernel 代码保存在这个路径
from ..triton_kernel.mhc_pre_op import *


def mhc_pre_op_fwd_kernel_call(
    x: torch.Tensor,
    h_res_in: torch.Tensor,
    h_pre_in: torch.Tensor,
    n: int,
    num_iters: int = 20,
    eps: float = 1e-8,
):
    """
    Triton Kernel Launcher for mHC Pre-Op Forward
    """
    # 1. 形状检查与准备
    B, T, _, C = x.shape
    Total_BT = B * T

    # 2. 确保内存连续 (Triton 性能关键)
    x = x.contiguous()
    h_res_in = h_res_in.contiguous()
    h_pre_in = h_pre_in.contiguous()

    # 3. 准备输出张量
    # x_layer_in: [B, T, C] (BF16)
    out = torch.empty((B, T, C), device=x.device, dtype=x.dtype)
    # H_res_out: [B, T, n, n] (FP32)
    H_res_out = torch.empty((B, T, n, n), device=x.device, dtype=torch.float32)

    # 4. 展平视图 (View as [Total_BT, ...])
    # 注意：stride 的获取必须基于传入的张量，而不是 view 后的
    # 但由于我们做了 contiguous，view 后的 stride 也是标准的

    # 5. Grid 计算函数
    # Grid: (Total_BT // BLOCK_BT, C // BLOCK_C)
    grid = lambda META: (
        triton.cdiv(Total_BT, META["BLOCK_BT"]),
        triton.cdiv(C, META["BLOCK_C"]),
    )

    # 6. 启动 Kernel
    sinkhorn_aggregate_fused_kernel[grid](
        # --- 指针 ---
        x,
        h_res_in,
        h_pre_in,
        out,
        H_res_out,
        # --- 常量 (constexpr) ---
        Total_BT_CONST=Total_BT,
        NSIZE=n,
        CSIZE=C,
        NUM_ITERS=num_iters,
        EPS=eps,
        # --- 步幅 (View as [Total_BT, ...]) ---
        # x: [BT, n, C] -> stride(0) 是 B维度stride, stride(1) 是 T维度stride
        # 对于 flatten 后的 [BT, n, C]，stride_bt 就是 n*C (若连续)
        # 最稳妥的方式是用 reshape 后的 stride
        stride_x_bt=x.view(Total_BT, n, C).stride(0),
        stride_x_n=x.stride(2),
        stride_x_c=x.stride(3),
        stride_h_res_in_bt=h_res_in.view(Total_BT, n, n).stride(0),
        stride_h_res_in_n1=h_res_in.stride(2),
        stride_h_res_in_n2=h_res_in.stride(3),
        stride_h_pre_in_bt=h_pre_in.view(Total_BT, n).stride(0),
        stride_h_pre_in_n=h_pre_in.stride(2),
        stride_out_bt=out.view(Total_BT, C).stride(0),
        stride_out_c=out.stride(2),
        stride_Hr_out_bt=H_res_out.view(Total_BT, n, n).stride(0),
        stride_Hr_out_n1=H_res_out.stride(2),
        stride_Hr_out_n2=H_res_out.stride(3),
    )

    return out, H_res_out


def mhc_pre_op_bwd_kernel_call(grad_out, grad_H_res, x, h_res, h_pre, n, iters, eps):
    B, T, _, C = x.shape
    BT = B * T
    gx = torch.empty_like(x)
    gh_res = torch.empty_like(h_res)
    gh_pre = torch.empty_like(h_pre)
    grid = (BT, 1)  # 消除原子操作的关键：Grid Y = 1

    sinkhorn_aggregate_bwd_kernel[grid](
        grad_out,
        grad_H_res,
        x,
        h_res,
        h_pre,
        gx,
        gh_res,
        gh_pre,
        TOTAL_BT_CONST=BT,
        NSIZE=n,
        CHANNEL_SIZE=C,
        NUM_ITERS=iters,
        EPS=eps,
        # 步幅映射 (与 Kernel 内部签名一一对应)
        stride_gout_bt=grad_out.view(BT, C).stride(0),
        stride_gout_c=grad_out.stride(2),
        stride_gH_bt=grad_H_res.view(BT, n, n).stride(0),
        stride_gH_n1=grad_H_res.stride(2),
        stride_gH_n2=grad_H_res.stride(3),
        stride_x_bt=x.view(BT, n, C).stride(0),
        stride_x_n=x.stride(2),
        stride_x_c=x.stride(3),
        stride_h_res_bt=h_res.view(BT, n, n).stride(0),
        stride_h_res_n1=h_res.stride(2),
        stride_h_res_n2=h_res.stride(3),
        stride_h_pre_bt=h_pre.view(BT, n).stride(0),
        stride_h_pre_n=h_pre.stride(2),
        stride_gx_bt=gx.view(BT, n, C).stride(0),
        stride_gx_n=gx.stride(2),
        stride_gx_c=gx.stride(3),
        stride_gh_res_bt=gh_res.view(BT, n, n).stride(0),
        stride_gh_res_n1=gh_res.stride(2),
        stride_gh_res_n2=gh_res.stride(3),
        stride_gh_pre_bt=gh_pre.view(BT, n).stride(0),
        stride_gh_pre_n=gh_pre.stride(2),
    )
    return gx, gh_res, gh_pre


class MHCFusedPreOp(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, h_res_in, h_pre_in, n, num_iters, eps):
        out, H_res_out = mhc_pre_op_fwd_kernel_call(
            x, h_res_in, h_pre_in, n, num_iters, eps
        )
        ctx.save_for_backward(x, h_res_in, h_pre_in)
        ctx.params = (n, num_iters, eps)
        return out, H_res_out

    @staticmethod
    def backward(ctx, grad_out, grad_H_res):
        x, h_res, h_pre = ctx.saved_tensors
        n, iters, eps = ctx.params
        if grad_out is None:
            grad_out = torch.zeros_like(x[:, :, 0])  # dummy
        if grad_H_res is None:
            grad_H_res = torch.zeros_like(h_res)
        gx, gh_res, gh_pre = mhc_pre_op_bwd_kernel_call(
            grad_out, grad_H_res, x, h_res, h_pre, n, iters, eps
        )
        return gx, gh_res, gh_pre, None, None, None


def mhc_pre_op_fused(
    x: torch.Tensor,
    h_res_in: torch.Tensor,
    h_pre_in: torch.Tensor,
    num_iters: int = 20,
    eps: float = 1e-8,
):
    """
    Triton 加速版 mHC Pre-Op (Fused Sinkhorn + Stream Aggregate)

    参数:
        x: [B, T, n, C] (BF16)
        h_res_in: [B, T, n, n] (FP32) - 未归一化的残差矩阵
        h_pre_in: [B, T, n] (FP32) - 未激活的聚合权重

    返回:
        out: [B, T, C] (BF16) - 聚合后的层输入
        H_res_out: [B, T, n, n] (FP32) - 双随机残差矩阵
    """
    n = x.shape[-2]
    return MHCFusedPreOp.apply(x, h_res_in, h_pre_in, n, num_iters, eps)



import torch
from ..triton_kernel.mhc_post_op import *


def mhc_post_op_forward(
    layer_out: torch.tensor,
    x_expanded: torch.tensor,
    h_post_raw: torch.tensor,
    H_res: torch.tensor,
) -> torch.tensor:
    batch, time, NSIZE, channel = x_expanded.shape
    total_batch_time = batch * time

    # 内存连续化处理
    x_v = x_expanded.reshape(-1, NSIZE, channel).contiguous()
    h_v = h_post_raw.reshape(-1, NSIZE).contiguous()
    H_v = H_res.reshape(-1, NSIZE, NSIZE).contiguous()
    l_v = layer_out.reshape(-1, channel).contiguous()

    output_tensor = torch.empty_like(x_v)

    grid = lambda META: (total_batch_time, triton.cdiv(channel, META["BLOCK_CHANNEL"]))

    mhc_fused_forward_kernel[grid](
        x_v,
        h_v,
        H_v,
        l_v,
        output_tensor,
        stride_output_batch_time=output_tensor.stride(0),
        stride_output_n_size=output_tensor.stride(1),
        stride_output_channel=output_tensor.stride(2),
        stride_x_batch_time=x_v.stride(0),
        stride_x_n_size=x_v.stride(1),
        stride_x_channel=x_v.stride(2),
        stride_h_batch_time=h_v.stride(0),
        stride_h_n_size=h_v.stride(1),
        stride_H_batch_time=H_v.stride(0),
        stride_H_n_size_1=H_v.stride(1),
        stride_H_n_size_2=H_v.stride(2),
        stride_layer_out_batch_time=l_v.stride(0),
        stride_layer_out_channel=l_v.stride(1),
        # Constants
        CHANNEL_SIZE=channel,
        NSIZE=NSIZE,
    )

    return output_tensor.view(batch, time, NSIZE, channel)


def mhc_post_op_backward(
    grad_output: torch.Tensor,
    layer_out: torch.Tensor,
    x_expanded: torch.Tensor,
    h_post_raw: torch.Tensor,
    H_res: torch.Tensor,
):
    B, T, n, C = x_expanded.shape
    total_bt = B * T
    device = x_expanded.device

    # 1. 连续化
    x_v = x_expanded.reshape(-1, n, C).contiguous()
    h_v = h_post_raw.reshape(-1, n).contiguous()
    H_v = H_res.reshape(-1, n, n).contiguous()
    l_v = layer_out.reshape(-1, C).contiguous()
    g_out_v = grad_output.reshape(-1, n, C).contiguous()

    # 2. 准备输出 (不使用 Workspace)
    grad_x = torch.empty_like(x_v)
    grad_l = torch.empty_like(l_v)
    # 规约结果使用 FP32 保证精度，最后再转
    grad_h = torch.empty_like(h_v, dtype=torch.float32)
    grad_H = torch.empty_like(H_v, dtype=torch.float32)

    # 3. 启动 Kernel: Grid Y 设为 1，强迫单个 Program 处理整行
    grid = (total_bt, 1)

    mhc_fused_backward_kernel[grid](
        x_v,
        h_v,
        H_v,
        l_v,
        g_out_v,
        grad_x,
        grad_h,
        grad_H,
        grad_l,
        # Strides
        stride_x_bt=x_v.stride(0),
        stride_x_n=x_v.stride(1),
        stride_x_c=x_v.stride(2),
        stride_h_bt=h_v.stride(0),
        stride_h_n=h_v.stride(1),
        stride_H_bt=H_v.stride(0),
        stride_H_n1=H_v.stride(1),
        stride_H_n2=H_v.stride(2),
        stride_l_bt=l_v.stride(0),
        stride_l_c=l_v.stride(1),
        stride_g_bt=g_out_v.stride(0),
        stride_g_n=g_out_v.stride(1),
        stride_g_c=g_out_v.stride(2),
        stride_gx_bt=grad_x.stride(0),
        stride_gx_n=grad_x.stride(1),
        stride_gx_c=grad_x.stride(2),
        stride_gl_bt=grad_l.stride(0),
        stride_gl_c=grad_l.stride(1),
        stride_gh_bt=grad_h.stride(0),
        stride_gh_n=grad_h.stride(1),
        stride_gH_bt=grad_H.stride(0),
        stride_gH_n1=grad_H.stride(1),
        stride_gH_n2=grad_H.stride(2),
        # Constants
        CHANNEL_SIZE=C,
        NSIZE=n,
    )

    return (
        grad_l.view(B, T, C),
        grad_x.view(B, T, n, C),
        grad_h.view(B, T, n).to(h_post_raw.dtype),
        grad_H.view(B, T, n, n).to(H_res.dtype),
    )


class MHCPostOpFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, layer_out, x_expanded, h_post_raw, H_res):
        # 保存张量以供反向传播使用
        ctx.save_for_backward(layer_out, x_expanded, h_post_raw, H_res)
        return mhc_post_op_forward(layer_out, x_expanded, h_post_raw, H_res)

    @staticmethod
    def backward(ctx, grad_output):
        # 获取 forward 中保存的张量
        layer_out, x_expanded, h_post_raw, H_res = ctx.saved_tensors
        # 计算梯度
        grads = mhc_post_op_backward(
            grad_output, layer_out, x_expanded, h_post_raw, H_res
        )
        return grads


def mhc_post_op(
    layer_out: torch.Tensor,
    x_expanded: torch.Tensor,
    h_post_raw: torch.Tensor,
    H_res: torch.Tensor,
) -> torch.Tensor:
    """
    Multi-Head Control (MHC) Post-Operation.

    该算子实现了 RWKV 模型中的 MHC 后处理融合逻辑，包括：
    1. 门控计算: w = 2 * sigmoid(h_post_raw)
    2. 通道分配: x_delta = w * layer_out
    3. 矩阵混合: x_mixed = H_res @ x_expanded
    4. 融合输出: Out = x_mixed + x_delta

    使用 Triton 实现全融合 Kernel，相比原生 PyTorch 具有更低的显存带宽占用和更高的执行效率。

    参数:
        layer_out (torch.Tensor): 形状为 [B, T, C]，bfloat16 类型。
        x_expanded (torch.Tensor): 形状为 [B, T, n, C]，bfloat16 类型。
        h_post_raw (torch.Tensor): 形状为 [B, T, n]，float32 或 bfloat16 类型。
        H_res (torch.Tensor): 形状为 [B, T, n, n]，float32 或 bfloat16 类型。

    返回:
        torch.Tensor: 形状为 [B, T, n, C]，bfloat16 类型。

    形状说明:
        B: Batch Size
        T: Sequence Length
        C: Channel Size (Hidden Dimension)
        n: Head Size (State Dimension, 通常为 4 或 8)
    """
    return MHCPostOpFunction.apply(
        layer_out.to(torch.bfloat16),
        x_expanded.to(torch.bfloat16),
        h_post_raw.to(torch.float32),
        H_res.to(torch.float32),
    )



import keras
from distutils.util import strtobool
import os
from keras import ops


def transpose_head(x, head_first):
    if head_first:
        return ops.transpose(x, (0, 2, 1, 3))
    else:
        return x


def get_generalized_delta_rule(HEAD_SIZE=64, KERNEL_TYPE="native"):
    assert HEAD_SIZE % 4 == 0
    from .native_keras_op import generalized_delta_rule

    if keras.config.backend() == "jax":
        import jax

        if jax.devices()[0].platform == "gpu":
            if KERNEL_TYPE == "cuda":
                from .jax_cuda_kernel.wkv7_jax import get_jax_generalized_delta_rule

                return get_jax_generalized_delta_rule(HEAD_SIZE)
            elif KERNEL_TYPE == "triton":
                from .jax_triton_kernel import generalized_delta_rule as jax_kernel

                return jax_kernel, generalized_delta_rule
    elif keras.config.backend() == "torch":
        import torch

        if torch.cuda.is_available():
            if KERNEL_TYPE == "cuda":
                from .torch_cuda_kernel.wkv7_torch import (
                    get_torch_generalized_delta_rule,
                )

                return get_torch_generalized_delta_rule(HEAD_SIZE)
            elif KERNEL_TYPE == "triton":
                from .torch_triton_kernel import generalized_delta_rule as triton_kernel

                return triton_kernel, generalized_delta_rule

    return generalized_delta_rule, generalized_delta_rule


def get_rnn_generalized_delta_rule(HEAD_SIZE=64, KERNEL_TYPE="native"):
    assert HEAD_SIZE % 4 == 0
    from .native_keras_op import generalized_delta_rule

    if KERNEL_TYPE == "cuda":
        if keras.config.backend() == "jax":
            import jax

            if jax.devices()[0].platform == "gpu":
                from .jax_cuda_kernel_single.wkv7_single_step_jax import (
                    get_jax_generalized_delta_rule_single_step,
                )

                return get_jax_generalized_delta_rule_single_step(HEAD_SIZE)
        elif keras.config.backend() == "torch":
            import torch

            if torch.cuda.is_available():
                from .torch_cuda_kernel_single.wkv7_single_step_torch import (
                    get_torch_generalized_delta_rule_single_step,
                )

                return get_torch_generalized_delta_rule_single_step(HEAD_SIZE)
    return generalized_delta_rule
